from codecs import getreader
from io import BytesIO, StringIO

from rdflib import Literal, URIRef
from rdflib.graph import Graph

rdf = """@prefix skos:
<http://www.w3.org/2004/02/skos/core#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix : <http://www.test.org/#> .

:world rdf:type skos:Concept;
    skos:prefLabel "World"@en.
:africa rdf:type skos:Concept;
    skos:prefLabel "Africa"@en;
    skos:broaderTransitive :world.
:CI rdf:type skos:Concept;
    skos:prefLabel "C\u00f4te d'Ivoire"@fr;
    skos:broaderTransitive :africa.
"""


rdf_utf8 = rdf.encode("utf-8")

rdf_reader = getreader("utf-8")(BytesIO(rdf.encode("utf-8")))


def test_a():
    """Test reading N3 from a unicode objects as data"""
    g = Graph()
    g.parse(data=rdf, format="n3")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


def test_b():
    """Test reading N3 from a utf8 encoded string as data"""
    g = Graph()
    g.parse(data=rdf_utf8, format="n3")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


def test_c():
    """Test reading N3 from a codecs.StreamReader, outputting unicode"""
    g = Graph()
    #    rdf_reader.seek(0)
    g.parse(source=rdf_reader, format="n3")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


def test_d():
    """Test reading N3 from a StringIO over the unicode object"""
    g = Graph()
    g.parse(source=StringIO(rdf), format="n3")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


def test_e():
    """Test reading N3 from a BytesIO over the string object"""
    g = Graph()
    g.parse(source=BytesIO(rdf_utf8), format="n3")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


# this is unicode
rdfxml = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
   xmlns:skos="http://www.w3.org/2004/02/skos/core#"
>
  <rdf:Description rdf:about="http://www.test.org/#CI">
    <rdf:type rdf:resource="http://www.w3.org/2004/02/skos/core#Concept"/>
    <skos:prefLabel xml:lang="fr">C\u00f4te d\'Ivoire</skos:prefLabel>
    <skos:broaderTransitive rdf:resource="http://www.test.org/#africa"/>
  </rdf:Description>
</rdf:RDF>
"""

# this is a str
rdfxml_utf8 = rdfxml.encode("utf-8")

rdfxml_reader = getreader("utf-8")(BytesIO(rdfxml.encode("utf-8")))


def test_xml_a():
    """Test reading XML from a unicode object as data"""
    g = Graph()
    g.parse(data=rdfxml, format="xml")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


def test_xml_b():
    """Test reading XML from a utf8 encoded string object as data"""
    g = Graph()
    g.parse(data=rdfxml_utf8, format="xml")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")


# The following two cases are currently not supported by Graph.parse
# def test_xml_c():
#     """Test reading XML from a codecs.StreamReader, outputting unicode"""
#     g = Graph()
#     g.parse(source=rdfxml_reader, format='xml')
#     v = g.value(subject=URIRef("http://www.test.org/#CI"), predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"))
#     assert v==Literal(u"C\u00f4te d'Ivoire", lang='fr')

# def test_xml_d():
#     """Test reading XML from a BytesIO created from unicode object"""
#     g = Graph()
#     g.parse(source=BytesIO(rdfxml), format='xml')
#     v = g.value(subject=URIRef("http://www.test.org/#CI"), predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"))
#     assert v==Literal(u"C\u00f4te d'Ivoire", lang='fr')


def test_xml_e():
    """Test reading XML from a BytesIO created from utf8 encoded string"""
    g = Graph()
    g.parse(source=BytesIO(rdfxml_utf8), format="xml")
    v = g.value(
        subject=URIRef("http://www.test.org/#CI"),
        predicate=URIRef("http://www.w3.org/2004/02/skos/core#prefLabel"),
    )
    assert v == Literal("C\u00f4te d'Ivoire", lang="fr")



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class FOAF(DefinedNamespace):
    """
    Friend of a Friend (FOAF) vocabulary

    The Friend of a Friend (FOAF) RDF vocabulary, described using W3C RDF Schema and the Web Ontology Language.

    Generated from: http://xmlns.com/foaf/spec/index.rdf
    Date: 2020-05-26 14:20:01.597998

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    account: URIRef  # Indicates an account held by this agent.
    accountName: (
        URIRef  # Indicates the name (identifier) associated with this online account.
    )
    accountServiceHomepage: (
        URIRef  # Indicates a homepage of the service provide for this online account.
    )
    age: URIRef  # The age in years of some agent.
    based_near: URIRef  # A location that something is based near, for some broadly human notion of near.
    birthday: URIRef  # The birthday of this Agent, represented in mm-dd string form, eg. '12-31'.
    currentProject: URIRef  # A current project this person works on.
    depiction: URIRef  # A depiction of some thing.
    depicts: URIRef  # A thing depicted in this representation.
    dnaChecksum: URIRef  # A checksum for the DNA of some thing. Joke.
    familyName: URIRef  # The family name of some person.
    family_name: URIRef  # The family name of some person.
    firstName: URIRef  # The first name of a person.
    focus: URIRef  # The underlying or 'focal' entity associated with some SKOS-described concept.
    fundedBy: URIRef  # An organization funding a project or person.
    geekcode: URIRef  # A textual geekcode for this person, see http://www.geekcode.com/geek.html
    gender: URIRef  # The gender of this Agent (typically but not necessarily 'male' or 'female').
    givenName: URIRef  # The given name of some person.
    givenname: URIRef  # The given name of some person.
    holdsAccount: URIRef  # Indicates an account held by this agent.
    img: URIRef  # An image that can be used to represent some thing (ie. those depictions which are particularly representative of something, eg. one's photo on a homepage).
    interest: URIRef  # A page about a topic of interest to this person.
    knows: URIRef  # A person known by this person (indicating some level of reciprocated interaction between the parties).
    lastName: URIRef  # The last name of a person.
    made: URIRef  # Something that was made by this agent.
    maker: URIRef  # An agent that  made this thing.
    member: URIRef  # Indicates a member of a Group
    membershipClass: (
        URIRef  # Indicates the class of individuals that are a member of a Group
    )
    myersBriggs: URIRef  # A Myers Briggs (MBTI) personality classification.
    name: URIRef  # A name for some thing.
    nick: URIRef  # A short informal nickname characterising an agent (includes login identifiers, IRC and other chat nicknames).
    page: URIRef  # A page or document about this thing.
    pastProject: URIRef  # A project this person has previously worked on.
    phone: URIRef  # A phone,  specified using fully qualified tel: URI scheme (refs: http://www.w3.org/Addressing/schemes.html#tel).
    plan: URIRef  # A .plan comment, in the tradition of finger and '.plan' files.
    primaryTopic: URIRef  # The primary topic of some page or document.
    publications: URIRef  # A link to the publications of this person.
    schoolHomepage: URIRef  # A homepage of a school attended by the person.
    sha1: URIRef  # A sha1sum hash, in hex.
    skypeID: URIRef  # A Skype ID
    status: URIRef  # A string expressing what the user is happy for the general public (normally) to know about their current activity.
    surname: URIRef  # The surname of some person.
    theme: URIRef  # A theme.
    thumbnail: URIRef  # A derived thumbnail image.
    tipjar: URIRef  # A tipjar document for this agent, describing means for payment and reward.
    title: URIRef  # Title (Mr, Mrs, Ms, Dr. etc)
    topic: URIRef  # A topic of some page or document.
    topic_interest: URIRef  # A thing of interest to this person.
    workInfoHomepage: URIRef  # A work info homepage of some person; a page about their work for some organization.
    workplaceHomepage: URIRef  # A workplace homepage of some person; the homepage of an organization they work for.

    # http://www.w3.org/2000/01/rdf-schema#Class
    Agent: URIRef  # An agent (eg. person, group, software or physical artifact).
    Document: URIRef  # A document.
    Group: URIRef  # A class of Agents.
    Image: URIRef  # An image.
    LabelProperty: URIRef  # A foaf:LabelProperty is any RDF property with textual values that serve as labels.
    OnlineAccount: URIRef  # An online account.
    OnlineChatAccount: URIRef  # An online chat account.
    OnlineEcommerceAccount: URIRef  # An online e-commerce account.
    OnlineGamingAccount: URIRef  # An online gaming account.
    Organization: URIRef  # An organization.
    Person: URIRef  # A person.
    PersonalProfileDocument: URIRef  # A personal profile RDF document.
    Project: URIRef  # A project (a collective endeavour of some kind).

    # http://www.w3.org/2002/07/owl#InverseFunctionalProperty
    aimChatID: URIRef  # An AIM chat ID
    homepage: URIRef  # A homepage for some thing.
    icqChatID: URIRef  # An ICQ chat ID
    isPrimaryTopicOf: URIRef  # A document that this thing is the primary topic of.
    jabberID: URIRef  # A jabber ID for something.
    logo: URIRef  # A logo representing some thing.
    mbox: URIRef  # A  personal mailbox, ie. an Internet mailbox associated with exactly one owner, the first owner of this mailbox. This is a 'static inverse functional property', in that  there is (across time and change) at most one individual that ever has any particular value for foaf:mbox.
    mbox_sha1sum: URIRef  # The sha1sum of the URI of an Internet mailbox associated with exactly one owner, the  first owner of the mailbox.
    msnChatID: URIRef  # An MSN chat ID
    openid: URIRef  # An OpenID for an Agent.
    weblog: URIRef  # A weblog of some thing (whether person, group, company etc.).
    yahooChatID: URIRef  # A Yahoo chat ID

    _NS = Namespace("http://xmlns.com/foaf/0.1/")



from __future__ import annotations

import warnings
from typing import IO, Optional
from uuid import uuid4

from rdflib import Dataset
from rdflib.plugins.serializers.nquads import _nq_row
from rdflib.plugins.serializers.nt import _nt_row
from rdflib.serializer import Serializer

add_remove_methods = {"add": "A", "remove": "D"}


class PatchSerializer(Serializer):
    """
    Creates an RDF patch file to add and remove triples/quads.
    Can either:
    - Create an add or delete patch for a single Dataset.
    - Create a patch to represent the difference between two Datasets.
    """

    def __init__(
        self,
        store: Dataset,
    ):
        self.store: Dataset = store
        super().__init__(store)

    def serialize(
        self,
        stream: IO[bytes],
        base: Optional[str] = None,
        encoding: Optional[str] = None,
        **kwargs,
    ):
        """
        Serialize the store to the given stream.
        :param stream: The stream to serialize to.
        :param base: The base URI to use for the serialization.
        :param encoding: The encoding to use for the serialization.
        :param kwargs: Additional keyword arguments.
        Supported keyword arguments:
        - operation: The operation to perform. Either 'add' or 'remove'.
        - target: The target Dataset to compare against.
        NB: Only one of 'operation' or 'target' should be provided.
        - header_id: The header ID to use.
        - header_prev: The previous header ID to use.
        """
        operation = kwargs.get("operation")
        target = kwargs.get("target")
        header_id = kwargs.get("header_id")
        header_prev = kwargs.get("header_prev")
        if not header_id:
            header_id = f"uuid:{uuid4()}"
        encoding = self.encoding
        if base is not None:
            warnings.warn("PatchSerializer does not support base.")
        if encoding is not None and encoding.lower() != self.encoding.lower():
            warnings.warn(
                "PatchSerializer does not use custom encoding. "
                f"Given encoding was: {encoding}"
            )

        def write_header():
            stream.write(f"H id <{header_id}> .\n".encode(encoding, "replace"))
            if header_prev:
                stream.write(f"H prev <{header_prev}>\n".encode(encoding, "replace"))
            stream.write("TX .\n".encode(encoding, "replace"))

        def write_triples(contexts, op_code, use_passed_contexts=False):
            for context in contexts:
                if not use_passed_contexts:
                    context = self.store.get_context(context.identifier)
                for triple in context:
                    stream.write(
                        self._patch_row(triple, context.identifier, op_code).encode(
                            encoding, "replace"
                        )
                    )

        if operation:
            assert operation in add_remove_methods, f"Invalid operation: {operation}"
        elif not target:
            # No operation specified and no target specified
            # Fall back to default operation of "add" to prevent a no-op
            operation = "add"
        write_header()
        if operation:
            operation_code = add_remove_methods.get(operation)
            write_triples(self.store.contexts(), operation_code)
        elif target:
            to_add, to_remove = self._diff(target)
            write_triples(to_add.contexts(), "A", use_passed_contexts=True)
            write_triples(to_remove.contexts(), "D", use_passed_contexts=True)

        stream.write("TC .\n".encode(encoding, "replace"))

    def _diff(self, target):
        rows_to_add = target - self.store
        rows_to_remove = self.store - target
        return rows_to_add, rows_to_remove

    def _patch_row(self, triple, context_id, operation):
        if context_id == self.store.default_context.identifier:
            return f"{operation} {_nt_row(triple)}"
        else:
            return f"{operation} {_nq_row(triple, context_id)}"



import logging
import textwrap
import xml.dom.minidom
from typing import Callable, Sequence

import pytest

import rdflib
from rdflib import RDF, Literal

try:
    import html5rdf  # noqa: F401

    have_html5rdf = True
except ImportError:
    have_html5rdf = False


def testPythonRoundtrip():  # noqa: N802
    l1 = Literal("<msg>hello</msg>", datatype=RDF.XMLLiteral)
    assert l1.value is not None, "xml must have been parsed"
    assert l1.datatype == RDF.XMLLiteral, "literal must have right datatype"

    l2 = Literal("<msg>good morning</msg>", datatype=RDF.XMLLiteral)
    assert l2.value is not None, "xml must have been parsed"
    assert not l1.eq(l2), "literals must NOT be equal"

    l3 = Literal(l1.value)
    assert l1.eq(l3), "roundtripped literals must be equal"
    assert l3.datatype == RDF.XMLLiteral, "literal must have right datatype"

    l4 = Literal("<msg >hello</msg>", datatype=RDF.XMLLiteral)
    assert l1 == l4
    assert l1.eq(l4)

    rdflib.NORMALIZE_LITERALS = False
    try:
        l4 = Literal("<msg >hello</msg>", datatype=RDF.XMLLiteral)
        assert l1 != l4
        assert l1.eq(l4)
    finally:
        rdflib.NORMALIZE_LITERALS = True


def testRDFXMLParse():  # noqa: N802
    rdfxml = """\
<rdf:RDF
    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
>

<rdf:Description rdf:about="http://example.org/">
    <dc:description rdf:parseType="Literal">
        <p xmlns="http://www.w3.org/1999/xhtml" xml:lang="en"></p>
    </dc:description>
</rdf:Description>

</rdf:RDF>"""

    g = rdflib.Graph()
    g.parse(data=rdfxml, format="xml")
    l1 = list(g)[0][2]
    assert l1.datatype == RDF.XMLLiteral


def graph():
    g = rdflib.Graph()
    g.add(
        (
            rdflib.URIRef("http://example.org/a"),
            rdflib.URIRef("http://example.org/p"),
            rdflib.Literal("<msg>hei</hei>", datatype=RDF.XMLLiteral),
        )
    )
    return g


def roundtrip(fmt):
    g1 = graph()
    l1 = list(g1)[0][2]
    g2 = rdflib.Graph()
    g2.parse(data=g1.serialize(format=fmt), format=fmt)
    l2 = list(g2)[0][2]
    assert l1.eq(l2)


def testRoundtrip():  # noqa: N802
    roundtrip("xml")
    roundtrip("n3")
    roundtrip("nt")


@pytest.mark.skipif(not have_html5rdf, reason="requires html5rdf")
def testHTML():  # noqa: N802
    l1 = Literal("<msg>hello</msg>", datatype=RDF.XMLLiteral)
    assert l1.value is not None, "xml must have been parsed"
    assert l1.datatype == RDF.XMLLiteral, "literal must have right datatype"

    l2 = Literal("<msg>hello</msg>", datatype=RDF.HTML)
    assert l2.value is not None, "xml must have been parsed"
    assert l2.datatype == RDF.HTML, "literal must have right datatype"

    l3 = Literal("<invalid", datatype=RDF.HTML)
    assert l3.value is None, "invalid html must not be parsed"
    assert l3.datatype == RDF.HTML, "literal must have right datatype"
    assert str(l3) == "<invalid", "invalid html must not be normalized"

    assert l1 != l2
    assert not l1.eq(l2)


@pytest.mark.parametrize(
    ("values", "equals"),
    [
        pytest.param(
            [
                lambda: Literal("<something />", datatype=RDF.XMLLiteral),
                lambda: Literal("<something/>", datatype=RDF.XMLLiteral),
            ],
            True,
        ),
        pytest.param(
            [
                lambda: Literal(
                    xml.dom.minidom.parseString(
                        textwrap.dedent(
                            """\
                    <!DOCTYPE example>
                    <something2/>
                    """
                        )
                    ),
                    datatype=RDF.XMLLiteral,
                ),
                lambda: Literal(
                    xml.dom.minidom.parseString(
                        textwrap.dedent(
                            """\
                    <!DOCTYPE example>
                    <something2 />
                    """
                        )
                    ),
                    datatype=RDF.XMLLiteral,
                ),
            ],
            True,
        ),
    ],
)
def test_eq(values: Sequence[Callable[[], Literal]], equals: bool) -> None:
    first_value = values[0]()
    logging.debug("first_value = \n%r", first_value)
    for value in values[1:]:
        current_value = value()
        logging.debug("current_value = \n%r", current_value)
        if equals:
            assert first_value.eq(
                current_value
            ), f"{current_value} must be equal to {first_value}"
        else:
            assert not first_value.eq(
                current_value
            ), f"{current_value} must be not be equal to {first_value}"



import pytest

from rdflib import OWL, BNode, Graph, Namespace, URIRef
from rdflib.extras.infixowl import (
    BooleanClass,
    Class,
    ComponentTerms,
    Individual,
    Property,
    some,
)

EXNS = Namespace("http://example.org/vocab/")
PZNS = Namespace(
    "http://www.co-ode.org/ontologies/pizza/2005/10/18/classified/pizza.owl#"
)


@pytest.fixture(scope="function")
def graph():
    g = Graph(identifier=EXNS.context0)
    g.bind("ex", EXNS)
    Individual.factoryGraph = g

    yield g

    del g


def test_componentterms_restriction():
    name = EXNS.Man
    assert isinstance(name, URIRef)

    r = (
        (Property(EXNS.someProp, baseType=OWL.DatatypeProperty))
        @ some
        @ (Class(EXNS.Foo))
    )

    r.value = BNode("foo")

    ct = ComponentTerms(r)
    assert str(list(ct)) in ["[Class: ns1:Foo ]", "[Class: ex:Foo ]"]


def test_componentterms_booleanclass(graph):
    fire = Class(EXNS.Fire)
    water = Class(EXNS.Water)

    id = EXNS.randoboolclass

    bc = BooleanClass(identifier=id, members=[fire, water], graph=graph)

    ct = ComponentTerms(bc)

    assert str(list(ct)) == "[Class: ex:Fire , Class: ex:Water ]"


def test_componentterms_booleanclass_bnodeid(graph):
    fire = Class(EXNS.Fire)
    water = Class(EXNS.Water)

    id = BNode("randoboolclass")

    bc = BooleanClass(identifier=id, members=[fire, water], graph=graph)

    ct = ComponentTerms(bc)

    assert str(list(ct)) == "[Class: ex:Fire , Class: ex:Water ]"



from __future__ import annotations

import logging
import re
from http.server import BaseHTTPRequestHandler
from urllib.error import HTTPError

import pytest

from rdflib import Graph
from test.data import TEST_DATA_DIR
from test.utils import GraphHelper
from test.utils.graph import cached_graph
from test.utils.http import (
    MOCK_HTTP_REQUEST_WILDCARD,
    MethodName,
    MockHTTPRequest,
    MockHTTPResponse,
    ctx_http_handler,
)
from test.utils.httpservermock import ServedBaseHTTPServerMock
from test.utils.namespace import EGDO
from test.utils.wildcard import URL_PARSE_RESULT_WILDCARD

"""
Test that correct content negotiation headers are passed
by graph.parse
"""


xmltestdoc = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns="http://example.org/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="http://example.org/a">
    <b rdf:resource="http://example.org/c"/>
  </rdf:Description>
</rdf:RDF>
"""

n3testdoc = """@prefix : <http://example.org/> .

:a :b :c .
"""

nttestdoc = "<http://example.org/a> <http://example.org/b> <http://example.org/c> .\n"

ttltestdoc = """@prefix : <http://example.org/> .

            :a :b :c .
            """

jsonldtestdoc = """
                [
                  {
                    "@id": "http://example.org/a",
                    "http://example.org/b": [
                      {
                        "@id": "http://example.org/c"
                      }
                    ]
                  }
                ]
                """


class ContentNegotiationHandler(BaseHTTPRequestHandler):
    def do_GET(self):  # noqa: N802
        self.send_response(200, "OK")
        # fun fun fun parsing accept header.

        acs = self.headers["Accept"].split(",")
        acq = [x.split(";") for x in acs if ";" in x]
        acn = [(x, "q=1") for x in acs if ";" not in x]
        acs = [(x[0].strip(), float(x[1].strip()[2:])) for x in acq + acn]
        ac = sorted(acs, key=lambda x: x[1])
        ct = ac[-1]

        if "application/rdf+xml" in ct:
            rct = "application/rdf+xml"
            content = xmltestdoc
        elif "text/n3" in ct:
            rct = "text/n3"
            content = n3testdoc
        elif "application/trig" in ct:
            rct = "application/trig"
            content = ttltestdoc
        elif "text/plain" in ct or "application/n-triples" in ct:
            rct = "text/plain"
            content = nttestdoc
        elif "application/ld+json" in ct:
            rct = "application/ld+json"
            content = jsonldtestdoc
        else:  # "text/turtle" in ct:
            rct = "text/turtle"
            content = ttltestdoc

        self.send_header("Content-type", rct)
        self.end_headers()
        self.wfile.write(content.encode("utf-8"))

    def log_message(self, *args):
        pass


class TestGraphHTTP:
    def test_content_negotiation(self) -> None:
        expected = Graph()
        expected.add((EGDO.a, EGDO.b, EGDO.c))
        expected_triples = GraphHelper.triple_set(expected)

        with ctx_http_handler(ContentNegotiationHandler) as server:
            (host, port) = server.server_address
            if isinstance(host, (bytes, bytearray)):
                host = host.decode("utf-8")
            url = f"http://{host}:{port}/foo"
            for format in ("xml", "n3", "nt"):
                graph = Graph()
                graph.parse(url, format=format)
                assert expected_triples == GraphHelper.triple_set(graph)

    def test_content_negotiation_no_format(self) -> None:
        expected = Graph()
        expected.add((EGDO.a, EGDO.b, EGDO.c))
        expected_triples = GraphHelper.triple_set(expected)

        with ctx_http_handler(ContentNegotiationHandler) as server:
            (host, port) = server.server_address
            if isinstance(host, (bytes, bytearray)):
                host = host.decode("utf-8")
            url = f"http://{host}:{port}/foo"
            graph = Graph()
            graph.parse(url)
            assert expected_triples == GraphHelper.triple_set(graph)

    def test_source(self) -> None:
        expected = Graph()
        expected.add((EGDO["a"], EGDO["b"], EGDO["c"]))
        expected_triples = GraphHelper.triple_set(expected)

        with ServedBaseHTTPServerMock() as httpmock:
            url = httpmock.url

            httpmock.responses[MethodName.GET].append(
                MockHTTPResponse(
                    200,
                    "OK",
                    f"<{EGDO['a']}> <{EGDO['b']}> <{EGDO['c']}>.".encode(),
                    {"Content-Type": ["text/turtle"]},
                )
            )
            graph = Graph()
            graph.parse(source=url)
            assert expected_triples == GraphHelper.triple_set(graph)

    def test_3xx(self) -> None:
        expected = Graph()
        expected.add((EGDO["a"], EGDO["b"], EGDO["c"]))
        expected_triples = GraphHelper.triple_set(expected)

        with ServedBaseHTTPServerMock() as httpmock:
            url = httpmock.url

            for idx in range(3):
                httpmock.responses[MethodName.GET].append(
                    MockHTTPResponse(
                        302,
                        "FOUND",
                        b"",
                        {"Location": [f"{url}/loc/302/{idx}"]},
                    )
                )
            for idx in range(3):
                httpmock.responses[MethodName.GET].append(
                    MockHTTPResponse(
                        303,
                        "See Other",
                        b"",
                        {"Location": [f"{url}/loc/303/{idx}"]},
                    )
                )
            for idx in range(3):
                httpmock.responses[MethodName.GET].append(
                    MockHTTPResponse(
                        308,
                        "Permanent Redirect",
                        b"",
                        {"Location": [f"{url}/loc/308/{idx}"]},
                    )
                )

            httpmock.responses[MethodName.GET].append(
                MockHTTPResponse(
                    200,
                    "OK",
                    f"<{EGDO['a']}> <{EGDO['b']}> <{EGDO['c']}>.".encode(),
                    {"Content-Type": ["text/turtle"]},
                )
            )

            graph = Graph()
            graph.parse(location=url, format="turtle")
            assert expected_triples == GraphHelper.triple_set(graph)

            httpmock.mocks[MethodName.GET].assert_called()
            assert len(httpmock.requests[MethodName.GET]) == 10
            for request in httpmock.requests[MethodName.GET]:
                # type error: Argument 2 to "match" has incompatible type "Optional[Any]"; expected "str"
                assert re.match(r"text/turtle", request.headers.get("Accept"))  # type: ignore[arg-type]

            request_paths = [
                request.path for request in httpmock.requests[MethodName.GET]
            ]
            assert request_paths == [
                "/",
                "/loc/302/0",
                "/loc/302/1",
                "/loc/302/2",
                "/loc/303/0",
                "/loc/303/1",
                "/loc/303/2",
                "/loc/308/0",
                "/loc/308/1",
                "/loc/308/2",
            ]

    def test_5xx(self):
        with ServedBaseHTTPServerMock() as httpmock:
            url = httpmock.url
            httpmock.responses[MethodName.GET].append(
                MockHTTPResponse(500, "Internal Server Error", b"", {})
            )

            graph = Graph()

            with pytest.raises(HTTPError) as raised:
                graph.parse(location=url, format="turtle")

            assert raised.value.code == 500


@pytest.mark.parametrize(
    ["url_suffix", "expected_request"],
    [
        (
            "/resource/Almería",
            MOCK_HTTP_REQUEST_WILDCARD._replace(
                path="/resource/Almer%C3%ADa",
                parsed_path=URL_PARSE_RESULT_WILDCARD._replace(
                    path="/resource/Almer%C3%ADa"
                ),
            ),
        ),
        (
            "/resource/Almería?foo=bar",
            MOCK_HTTP_REQUEST_WILDCARD._replace(
                parsed_path=URL_PARSE_RESULT_WILDCARD._replace(
                    path="/resource/Almer%C3%ADa"
                ),
                path_query={"foo": ["bar"]},
            ),
        ),
    ],
)
def test_iri_source(
    url_suffix: str,
    expected_request: MockHTTPRequest,
    function_httpmock: ServedBaseHTTPServerMock,
) -> None:
    diverse_triples_path = TEST_DATA_DIR / "variants/diverse_triples.ttl"

    function_httpmock.responses[MethodName.GET].append(
        MockHTTPResponse(
            200,
            "OK",
            diverse_triples_path.read_bytes(),
            {"Content-Type": ["text/turtle"]},
        )
    )
    g = Graph()
    g.parse(f"{function_httpmock.url}{url_suffix}")
    assert function_httpmock.call_count == 1
    GraphHelper.assert_triple_sets_equals(cached_graph((diverse_triples_path,)), g)
    assert len(g) > 1

    req = function_httpmock.requests[MethodName.GET].pop(0)
    logging.debug("req = %s", req)
    assert expected_request == req



"""\
N-Triples Parser
License: GPL 2, W3C, BSD, or MIT
Author: Sean B. Palmer, inamidst.com
"""

from __future__ import annotations

import codecs
import re
from io import BytesIO, StringIO, TextIOBase
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Match,
    MutableMapping,
    Optional,
    Pattern,
    TextIO,
    Union,
)

from rdflib.compat import _string_escape_map, decodeUnicodeEscape
from rdflib.exceptions import ParserError as ParseError
from rdflib.parser import InputSource, Parser
from rdflib.term import BNode as bNode
from rdflib.term import Literal, URIRef
from rdflib.term import URIRef as URI  # noqa: N814

if TYPE_CHECKING:
    import typing_extensions as te

    from rdflib.graph import Graph, _ObjectType, _PredicateType, _SubjectType

__all__ = [
    "unquote",
    "uriquote",
    "W3CNTriplesParser",
    "NTGraphSink",
    "NTParser",
    "DummySink",
]

uriref = r'<([^:]+:[^\s"<>]*)>'
literal = r'"([^"\\]*(?:\\.[^"\\]*)*)"'
litinfo = r"(?:@([a-zA-Z]+(?:-[a-zA-Z0-9]+)*)|\^\^" + uriref + r")?"

r_line = re.compile(r"([^\r\n]*)(?:\r\n|\r|\n)")
r_wspace = re.compile(r"[ \t]*")
r_wspaces = re.compile(r"[ \t]+")
r_tail = re.compile(r"[ \t]*\.[ \t]*(#.*)?")
r_uriref = re.compile(uriref)
r_nodeid = re.compile(r"_:([A-Za-z0-9_:]([-A-Za-z0-9_:\.]*[-A-Za-z0-9_:])?)")
r_literal = re.compile(literal + litinfo)

bufsiz = 2048
validate = False


class DummySink:
    def __init__(self):
        self.length = 0

    def triple(self, s, p, o):
        self.length += 1
        print(s, p, o)


r_safe = re.compile(r"([\x20\x21\x23-\x5B\x5D-\x7E]+)")
r_quot = re.compile(r"""\\([tbnrf"'\\])""")
r_uniquot = re.compile(r"\\u([0-9A-Fa-f]{4})|\\U([0-9A-Fa-f]{8})")


def unquote(s: str) -> str:
    """Unquote an N-Triples string."""
    if not validate:
        if isinstance(s, str):  # nquads
            s = decodeUnicodeEscape(s)
        else:
            s = s.decode("unicode-escape")  # type: ignore[unreachable]

        return s
    else:
        result = []
        while s:
            m = r_safe.match(s)
            if m:
                s = s[m.end() :]
                result.append(m.group(1))
                continue

            m = r_quot.match(s)
            if m:
                s = s[2:]
                result.append(_string_escape_map[m.group(1)])
                continue

            m = r_uniquot.match(s)
            if m:
                s = s[m.end() :]
                u, U = m.groups()  # noqa: N806
                codepoint = int(u or U, 16)
                if codepoint > 0x10FFFF:
                    raise ParseError("Disallowed codepoint: %08X" % codepoint)
                result.append(chr(codepoint))
            elif s.startswith("\\"):
                raise ParseError("Illegal escape at: %s..." % s[:10])
            else:
                raise ParseError("Illegal literal character: %r" % s[0])
        return "".join(result)


r_hibyte = re.compile(r"([\x80-\xFF])")


def uriquote(uri: str) -> str:
    if not validate:
        return uri
    else:
        return r_hibyte.sub(lambda m: "%%%02X" % ord(m.group(1)), uri)


_BNodeContextType = MutableMapping[str, bNode]


class W3CNTriplesParser:
    """An N-Triples Parser.
    This is a legacy-style Triples parser for NTriples provided by W3C
    Usage::

          p = W3CNTriplesParser(sink=MySink())
          sink = p.parse(f) # file; use parsestring for a string

    To define a context in which blank node identifiers refer to the same blank node
    across instances of NTriplesParser, pass the same dict as ``bnode_context`` to each
    instance. By default, a new blank node context is created for each instance of
    `W3CNTriplesParser`.
    """

    __slots__ = ("_bnode_ids", "sink", "buffer", "file", "line", "skolemize")

    def __init__(
        self,
        sink: Optional[Union[DummySink, NTGraphSink]] = None,
        bnode_context: Optional[_BNodeContextType] = None,
    ):
        self.skolemize = False

        if bnode_context is not None:
            self._bnode_ids = bnode_context
        else:
            self._bnode_ids = {}

        self.sink: Union[DummySink, NTGraphSink]
        if sink is not None:
            self.sink = sink
        else:
            self.sink = DummySink()

        self.buffer: Optional[str] = None
        self.file: Optional[Union[TextIO, codecs.StreamReader]] = None
        self.line: Optional[str] = ""

    def parse(
        self,
        f: Union[TextIO, IO[bytes], codecs.StreamReader],
        bnode_context: Optional[_BNodeContextType] = None,
        skolemize: bool = False,
    ) -> Union[DummySink, NTGraphSink]:
        """
        Parse f as an N-Triples file.

        :type f: :term:`file object`
        :param f: the N-Triples source
        :type bnode_context: `dict`, optional
        :param bnode_context: a dict mapping blank node identifiers (e.g., ``a`` in ``_:a``)
                              to `~rdflib.term.BNode` instances. An empty dict can be
                              passed in to define a distinct context for a given call to
                              `parse`.
        """

        if not hasattr(f, "read"):
            raise ParseError("Item to parse must be a file-like object.")

        if not hasattr(f, "encoding") and not hasattr(f, "charbuffer"):
            # someone still using a bytestream here?
            f = codecs.getreader("utf-8")(f)

        self.skolemize = skolemize
        self.file = f  # type: ignore[assignment]
        self.buffer = ""
        while True:
            self.line = self.readline()
            if self.line is None:
                break
            try:
                self.parseline(bnode_context=bnode_context)
            except ParseError:
                raise ParseError("Invalid line: {}".format(self.line))
        return self.sink

    def parsestring(self, s: Union[bytes, bytearray, str], **kwargs) -> None:
        """Parse s as an N-Triples string."""
        if not isinstance(s, (str, bytes, bytearray)):
            raise ParseError("Item to parse must be a string instance.")
        f: Union[codecs.StreamReader, StringIO]
        if isinstance(s, (bytes, bytearray)):
            f = codecs.getreader("utf-8")(BytesIO(s))
        else:
            f = StringIO(s)
        self.parse(f, **kwargs)

    def readline(self) -> Optional[str]:
        """Read an N-Triples line from buffered input."""
        # N-Triples lines end in either CRLF, CR, or LF
        # Therefore, we can't just use f.readline()
        if not self.buffer:
            # type error: Item "None" of "Union[TextIO, StreamReader, None]" has no attribute "read"
            buffer = self.file.read(bufsiz)  # type: ignore[union-attr]
            if not buffer:
                return None
            self.buffer = buffer

        while True:
            m = r_line.match(self.buffer)
            if m:  # the more likely prospect
                self.buffer = self.buffer[m.end() :]
                return m.group(1)
            else:
                # type error: Item "None" of "Union[TextIO, StreamReader, None]" has no attribute "read"
                buffer = self.file.read(bufsiz)  # type: ignore[union-attr]
                if not buffer and not self.buffer.isspace():
                    # Last line does not need to be terminated with a newline
                    buffer += "\n"
                elif not buffer:
                    return None
                self.buffer += buffer

    def parseline(self, bnode_context: Optional[_BNodeContextType] = None) -> None:
        self.eat(r_wspace)
        if (not self.line) or self.line.startswith("#"):
            return  # The line is empty or a comment

        subject = self.subject(bnode_context)
        self.eat(r_wspaces)

        predicate = self.predicate()
        self.eat(r_wspaces)

        object_ = self.object(bnode_context)
        self.eat(r_tail)

        if self.line:
            raise ParseError("Trailing garbage: {}".format(self.line))
        self.sink.triple(subject, predicate, object_)

    def peek(self, token: str) -> bool:
        return self.line.startswith(token)  # type: ignore[union-attr]

    def eat(self, pattern: Pattern[str]) -> Match[str]:
        m = pattern.match(self.line)  # type: ignore[arg-type]
        if not m:  # @@ Why can't we get the original pattern?
            # print(dir(pattern))
            # print repr(self.line), type(self.line)
            raise ParseError("Failed to eat %s at %s" % (pattern.pattern, self.line))
        self.line = self.line[m.end() :]  # type: ignore[index]
        return m

    def subject(self, bnode_context=None) -> Union[bNode, URIRef]:
        # @@ Consider using dictionary cases
        subj = self.uriref() or self.nodeid(bnode_context)
        if not subj:
            raise ParseError("Subject must be uriref or nodeID")
        return subj

    def predicate(self) -> Union[bNode, URIRef]:
        pred = self.uriref()
        if not pred:
            raise ParseError("Predicate must be uriref")
        return pred

    def object(
        self, bnode_context: Optional[_BNodeContextType] = None
    ) -> Union[URI, bNode, Literal]:
        objt = self.uriref() or self.nodeid(bnode_context) or self.literal()
        if objt is False:
            raise ParseError("Unrecognised object type")
        return objt

    def uriref(self) -> Union[te.Literal[False], URI]:
        if self.peek("<"):
            uri = self.eat(r_uriref).group(1)
            uri = unquote(uri)
            uri = uriquote(uri)
            return URI(uri)
        return False

    def nodeid(
        self, bnode_context: Optional[_BNodeContextType] = None
    ) -> Union[te.Literal[False], bNode, URI]:
        if self.peek("_"):
            if self.skolemize:
                bnode_id = self.eat(r_nodeid).group(1)
                return bNode(bnode_id).skolemize()

            else:
                # Fix for https://github.com/RDFLib/rdflib/issues/204
                if bnode_context is None:
                    bnode_context = self._bnode_ids
                bnode_id = self.eat(r_nodeid).group(1)
                new_id = bnode_context.get(bnode_id, None)
                if new_id is not None:
                    # Re-map to id specific to this doc
                    return bNode(new_id)
                else:
                    # Replace with freshly-generated document-specific BNode id
                    bnode = bNode()
                    # Store the mapping
                    bnode_context[bnode_id] = bnode
                    return bnode
        return False

    def literal(self) -> Union[te.Literal[False], Literal]:
        if self.peek('"'):
            lit, lang, dtype = self.eat(r_literal).groups()
            if lang:
                lang = lang
            else:
                lang = None
            if dtype:
                dtype = unquote(dtype)
                dtype = uriquote(dtype)
                dtype = URI(dtype)
            else:
                dtype = None
            if lang and dtype:
                raise ParseError("Can't have both a language and a datatype")
            lit = unquote(lit)
            return Literal(lit, lang, dtype)
        return False


class NTGraphSink:
    __slots__ = ("g",)

    def __init__(self, graph: Graph):
        self.g = graph

    def triple(self, s: _SubjectType, p: _PredicateType, o: _ObjectType) -> None:
        self.g.add((s, p, o))


class NTParser(Parser):
    """parser for the ntriples format, often stored with the .nt extension

    See http://www.w3.org/TR/rdf-testcases/#ntriples"""

    __slots__ = ()

    @classmethod
    def parse(cls, source: InputSource, sink: Graph, **kwargs: Any) -> None:
        """
        Parse the NT format

        :type source: `rdflib.parser.InputSource`
        :param source: the source of NT-formatted data
        :type sink: `rdflib.graph.Graph`
        :param sink: where to send parsed triples
        :param kwargs: Additional arguments to pass to `.W3CNTriplesParser.parse`
        """
        f: Union[TextIO, IO[bytes], codecs.StreamReader]
        f = source.getCharacterStream()
        if not f:
            b = source.getByteStream()
            # TextIOBase includes: StringIO and TextIOWrapper
            if isinstance(b, TextIOBase):
                # f is not really a ByteStream, but a CharacterStream
                f = b  # type: ignore[assignment]
            else:
                # since N-Triples 1.1 files can and should be utf-8 encoded
                f = codecs.getreader("utf-8")(b)
        parser = W3CNTriplesParser(NTGraphSink(sink))
        parser.parse(f, **kwargs)
        f.close()



from __future__ import annotations

import math
import sys
from typing import Set, Tuple

from rdflib import Graph, Literal
from rdflib.namespace import Namespace
from rdflib.plugins.sparql.processor import processUpdate
from rdflib.term import Node


def triple_set(graph: Graph) -> Set[Tuple[Node, Node, Node]]:
    return set(graph.triples((None, None, None)))


class TestSPARQLParser:
    def test_insert_recursionlimit(self) -> None:
        # These values are experimentally determined
        # to cause the RecursionError reported in
        # https://github.com/RDFLib/rdflib/issues/1336
        resource_count = math.ceil(sys.getrecursionlimit() / (33 - 3))
        self.do_insert(resource_count)

    def test_insert_large(self) -> None:
        self.do_insert(200)

    def do_insert(self, resource_count: int) -> None:
        EGV = Namespace("http://example.org/vocab#")  # noqa: N806
        EGI = Namespace("http://example.org/instance#")  # noqa: N806
        prop0, prop1, prop2 = EGV["prop0"], EGV["prop1"], EGV["prop2"]
        g0 = Graph()
        for index in range(resource_count):
            resource = EGI[f"resource{index}"]
            g0.add((resource, prop0, Literal(index)))
            g0.add((resource, prop1, Literal("example resource")))
            g0.add((resource, prop2, Literal(f"resource #{index}")))

        g0ntriples = g0.serialize(format="ntriples")
        g1 = Graph()

        assert triple_set(g0) != triple_set(g1)

        processUpdate(g1, f"INSERT DATA {{ {g0ntriples!s} }}")

        assert triple_set(g0) == triple_set(g1)



"""
This module contains type aliases that should only be used when type checking
as it would otherwise introduce a runtime dependency on `typing_extensions` for
older python versions which is not desirable.

This was made mainly to accommodate ``sphinx-autodoc-typehints`` which cannot
recognize type aliases from imported files if the type aliases are defined
inside ``if TYPE_CHECKING:``. So instead of placing the type aliases in normal
modules inside ``TYPE_CHECKING`` guards they are in this file which should only
be imported inside ``TYPE_CHECKING`` guards.

.. important::
    Things inside this module are not for use outside of RDFLib
    and this module is not part the the RDFLib public API.
"""

__all__ = [
    "_NamespaceSetString",
    "_MulPathMod",
]


from typing import Literal as PyLiteral

_NamespaceSetString = PyLiteral["core", "rdflib", "none"]
_MulPathMod = PyLiteral["*", "+", "?"]



# test for https://github.com/RDFLib/rdflib/issues/532

from rdflib import Graph


def test_issue532():
    data = """
    @base <http://purl.org/linkedpolitics/MembersOfParliament_background> .
    @prefix foaf: <http://xmlns.com/foaf/0.1/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    @prefix lpv: <vocabulary/> .
    @prefix NationalParty: <NationalParty/> .

    <EUmember_1026>
        a lpv:MemberOfParliament ;
        lpv:MEP_ID "1026" ;
        lpv:countryOfRepresentation <EUCountry_FR> ;
        lpv:dateOfBirth "1946-07-13"^^xsd:date ;
        lpv:politicalFunction [
            a lpv:PoliticalFunction ;
            lpv:beginning "1989-10-13"^^xsd:date ;
            lpv:end "1992-01-14"^^xsd:date ;
            lpv:institution NationalParty:sans_etiquette
        ] , [
            a lpv:PoliticalFunction ;
            lpv:beginning "2005-02-24"^^xsd:date ;
            lpv:end "2007-01-30"^^xsd:date ;
            lpv:institution NationalParty:Union_pour_la_democratie_francaise
        ] ;
        foaf:familyName "Bourlanges" ;
        foaf:givenName "Jean-Louis" .
    """

    g = Graph()
    g.parse(data=data, format="n3")

    getnewMeps = """
    PREFIX lpv: <http://purl.org/linkedpolitics/vocabulary/>
    prefix foaf: <http://xmlns.com/foaf/0.1/>
    prefix xsd: <http://www.w3.org/2001/XMLSchema#>

    SELECT DISTINCT ?name ?lastname WHERE {
      ?member a lpv:MemberOfParliament .
      ?member foaf:givenName ?name .
      ?member foaf:familyName ?lastname .
      ?member lpv:politicalFunction ?function .
      ?function lpv:beginning ?date .

      FILTER (?date >= "2004-06-20"^^xsd:date)
    }
    """  # noqa: N806

    result = list(g.query(getnewMeps))
    assert len(result) == 1



import pytest

from rdflib import BNode, Dataset, Graph
from rdflib.compare import isomorphic


@pytest.mark.parametrize(
    "data, data_format, expected_data, expected_data_format",
    [
        [
            """
        <urn:object> <urn:hasPart> _:internal-bnode-id-1 .
        _:internal-bnode-id-1 <urn:value> "..." .
        """,
            "ntriples",
            """
            <urn:object> <urn:hasPart> <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> .
            <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <urn:value> "..." .
        """,
            "ntriples",
        ]
    ],
)
def test_parse_with_skolemize_triples(
    data: str, data_format: str, expected_data: str, expected_data_format: str
):
    graph = Graph().parse(data=data, format=data_format, skolemize=True)
    assert len(graph)

    expected_graph = Graph().parse(data=expected_data, format=expected_data_format)
    assert len(expected_graph)

    assert isomorphic(graph, expected_graph)

    de_skolem_graph = graph.de_skolemize()
    expected_de_skolem_graph = expected_graph.de_skolemize()
    assert isomorphic(de_skolem_graph, expected_de_skolem_graph)


@pytest.mark.parametrize(
    "data, data_format, expected_data, expected_data_format, anonymous_graph_name",
    [
        [
            """
            <urn:object> <urn:hasPart> _:internal-bnode-id-1 _:graph-id .
            _:internal-bnode-id-1 <urn:value> "..." _:graph-id .
        """,
            "nquads",
            """
            <urn:object> <urn:hasPart> <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <https://rdflib.github.io/.well-known/genid/rdflib/graph-id>  .
            <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <urn:value> "..." <https://rdflib.github.io/.well-known/genid/rdflib/graph-id> .
        """,
            "nquads",
            "graph-id",
        ],
        [
            """
                ["urn:object", "urn:hasPart", "_:internal-bnode-id-1", "localId", "", "_:graph-id"]
                ["_:internal-bnode-id-1", "urn:value", "...", "http://www.w3.org/2001/XMLSchema#string", "", "_:graph-id"]
            """,
            "hext",
            """
                <urn:object> <urn:hasPart> <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <https://rdflib.github.io/.well-known/genid/rdflib/graph-id>  .
                <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <urn:value> "..."^^<http://www.w3.org/2001/XMLSchema#string> <https://rdflib.github.io/.well-known/genid/rdflib/graph-id> .
            """,
            "nquads",
            "graph-id",
        ],
        [
            """
                [
                    {
                        "@id": "_:graph-id",
                        "@graph": [
                            {
                                "@id": "urn:object",
                                "urn:hasPart": {
                                    "@id": "_:internal-bnode-id-1"
                                }
                            },
                            {
                                "@id": "_:internal-bnode-id-1",
                                "urn:value": "..."
                            }
                        ]
                    }
                ]
            """,
            "json-ld",
            """
                <urn:object> <urn:hasPart> <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <https://rdflib.github.io/.well-known/genid/rdflib/graph-id>  .
                <https://rdflib.github.io/.well-known/genid/rdflib/internal-bnode-id-1> <urn:value> "..." <https://rdflib.github.io/.well-known/genid/rdflib/graph-id> .
            """,
            "nquads",
            "graph-id",
        ],
    ],
)
def test_parse_with_skolemize_quads(
    data: str,
    data_format: str,
    expected_data: str,
    expected_data_format: str,
    anonymous_graph_name,
):
    ds = Dataset(default_union=True)
    ds.parse(data=data, format=data_format, skolemize=True)
    assert len(ds)

    expected_ds = Dataset(default_union=True)
    expected_ds.parse(data=expected_data, format=expected_data_format)
    assert len(expected_ds)

    graph_name = BNode(anonymous_graph_name)
    skolem_graph_name = graph_name.skolemize()

    skolem_graph = ds.graph(skolem_graph_name)
    expected_skolem_graph = expected_ds.graph(skolem_graph_name)
    assert len(skolem_graph)
    assert len(expected_skolem_graph)
    assert isomorphic(skolem_graph, expected_skolem_graph)
    assert isomorphic(skolem_graph.de_skolemize(), expected_skolem_graph.de_skolemize())

    # Note: Datasets must have default_union set to True, otherwise calling
    # de_skolemize returns an empty graph.
    assert isomorphic(ds.de_skolemize(), expected_ds.de_skolemize())

    # TODO: There's no way to roundtrip datasets with skolemization?



from __future__ import annotations

import urllib.request
from contextlib import contextmanager
from typing import Generator, Optional
from urllib.request import OpenerDirector, install_opener


@contextmanager
def context_urlopener(opener: OpenerDirector) -> Generator[OpenerDirector, None, None]:
    old_opener: Optional[OpenerDirector] = urllib.request._opener  # type: ignore[attr-defined]
    try:
        install_opener(opener)
        yield opener
    finally:
        install_opener(old_opener)  # type: ignore[arg-type]



import subprocess
import sys
from pathlib import Path

from rdflib.graph import Graph
from rdflib.tools.defined_namespace_creator import get_target_namespace_elements


def test_definednamespace_creator_qb():
    """
    Tests basic use of DefinedNamespace creator script using QB
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    qb_data_file = (
        Path(__file__).parent.parent / "data" / "defined_namespaces" / "qb.ttl"
    )
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {qb_data_file}...")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(qb_data_file),
            "http://purl.org/linked-data/cube#",
            "QB",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 0, "subprocess exited incorrectly"
    assert Path.is_file(Path("_QB.py")), "_QB.py file not created"
    has_ns = False
    has_test_class = False
    with open(Path("_QB.py")) as f:
        for line in f.readlines():
            if '_NS = Namespace("http://purl.org/linked-data/cube#")' in line:
                has_ns = True
            if (
                "Attachable: URIRef  # Abstract superclass for everything that can have attributes and dimensions"
                in line
            ):
                has_test_class = True
    assert has_ns, "_QB.py does not contain _NS"
    assert has_test_class, "_QB.py does not class Attachable"

    # cleanup
    Path.unlink(Path("_QB.py"))


def test_definednamespace_creator_fake():
    """
    Tests incorrect use of DefinedNamespace creator script -
    RDF file of unknonwn type
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    qb_data_file = (
        Path(__file__).parent.parent / "data" / "defined_namespaces" / "fake.xxx"
    )
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {qb_data_file}...(expected to fail)")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(qb_data_file),
            "http://purl.org/linked-data/cube#",
            "QB",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 1, "subprocess exited incorrectly (failure expected)"


def test_definednamespace_creator_bad_ns():
    """
    Tests incorrect use of DefinedNamespace creator script -
    supplied namespace doesn't end in # or /
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    qb_data_file = Path(__file__).parent.parent / "defined_namespaces" / "fake.xxx"
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {qb_data_file}...(expected to fail - bad NS given)")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(qb_data_file),
            "http://purl.org/linked-data/cube",
            "QB",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 1, "subprocess exited incorrectly (failure expected)"


def test_definednamespace_creator_multiple_comments():
    """
    Tests that only a single URIRef is declared, even when multiple
    rdfs:comments are linked to the resource.
    """

    definednamespace_script = (
        Path(__file__).parent.parent.parent
        / "rdflib"
        / "tools"
        / "defined_namespace_creator.py"
    )
    multiple_comments_data_file = (
        Path(__file__).parent.parent / "data" / "contrived" / "multiple-comments.ttl"
    )
    print("\n")
    print(f"Using {definednamespace_script}...")
    print(f"Testing {multiple_comments_data_file}...")
    completed = subprocess.run(
        [
            sys.executable,
            str(definednamespace_script),
            str(multiple_comments_data_file),
            "http://example.org/multiline-string-example#",
            "MULTILINESTRINGEXAMPLE",
        ],
        capture_output=True,
        text=True,
    )
    assert completed.returncode == 0, "subprocess exited incorrectly"
    assert Path.is_file(
        Path("_MULTILINESTRINGEXAMPLE.py")
    ), "_MULTILINESTRINGEXAMPLE.py file not created"

    some_class_count = 0
    with open(Path("_MULTILINESTRINGEXAMPLE.py")) as f:
        for line in f.readlines():
            if "SomeClass: URIRef" in line:
                some_class_count += 1

    assert (
        some_class_count == 1
    ), f"found {some_class_count} SomeClass definitions instead of 1."

    # cleanup
    Path.unlink(Path("_MULTILINESTRINGEXAMPLE.py"))


def test_get_target_namespace_elements(rdfs_graph: Graph) -> None:
    elements = get_target_namespace_elements(
        rdfs_graph, "http://www.w3.org/2000/01/rdf-schema#"
    )
    assert 3 == len(elements)
    assert 15 == len(elements[0])
    assert (
        "http://www.w3.org/2000/01/rdf-schema#Class",
        "The class of classes.",
    ) in elements[0]
    assert ("http://www.w3.org/2000/01/rdf-schema#", "") not in elements[0]
    assert 15 == len(elements[1])
    assert "    Class: URIRef  # The class of classes.\n" in elements[1]
    assert 0 == len(elements[2])



from __future__ import annotations

import logging
import os
from dataclasses import dataclass, field
from io import StringIO
from pathlib import Path
from typing import Collection, Tuple, Union, cast

import pytest
from _pytest.mark.structures import Mark, MarkDecorator, ParameterSet

import rdflib.plugins.sparql.algebra as algebra
import rdflib.plugins.sparql.parser as parser
from rdflib import Graph, Literal, URIRef, Variable
from rdflib.plugins.sparql.algebra import translateAlgebra
from test.data import TEST_DATA_DIR


@pytest.fixture
def data_path() -> Path:
    return TEST_DATA_DIR / "translate_algebra"


@dataclass
class AlgebraTest:
    id: str
    description: str
    marks: Union[MarkDecorator, Collection[Union[MarkDecorator, Mark]]] = field(
        default_factory=lambda: cast(Tuple[MarkDecorator], tuple())
    )

    @property
    def filename(self) -> str:
        return f"{self.id}.txt"

    def pytest_param(self) -> ParameterSet:
        return pytest.param(self, id=self.id, marks=self.marks)


def _format_query(query: str) -> str:
    buffer = StringIO()
    p = "{"
    q = "}"
    i = 0
    f = 1

    for e in query:
        if e in p:
            if not f:
                buffer.write("\n")
            buffer.write(" " * i + e)
            buffer.write("\n")
            i += 4
            f = 1
        elif e in q:
            if not f:
                buffer.write("\n")
            i -= 4
            f = 1
            buffer.write(" " * i + e)
            buffer.write("\n")
        else:
            if f:
                buffer.write(" " * i)
            buffer.write(e)
    return buffer.getvalue()


algebra_tests = [
    AlgebraTest(
        "test_functions__functional_forms",
        "Test if functional forms are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_functions__functional_forms_not_exists",
        "Test if the not exists form is properly translated into the query text.",
    ),
    AlgebraTest(
        "test_functions__functions_on_dates_and_time",
        "Test if functions on dates and time are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_functions__functions_on_numerics",
        "Test if functions on numerics are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_functions__functions_on_rdf_terms",
        "Test if functions on rdf terms are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_functions__functions_on_strings",
        "Test if functions on strings are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_functions__hash_functions",
        "Test if hash functions are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__aggregate_join",
        "Test if aggregate join including all aggregation functions"
        "are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__bgp",
        "Test if basic graph patterns are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__bnode_subject",
        "Test if a blank node subject gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__extend",
        'Test if "extend" (=Bind explicitly or implicitly in projection)'
        "gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__filter",
        "Test if filter gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__graph",
        'Test if "graph" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_graph_patterns__group",
        'Test if "group" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_graph_patterns__group_and_substr",
        'Test if a query with a variable that is used in the "GROUP BY" clause '
        "and in the SUBSTR function gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__group_and_nested_concat",
        "Test if a query with a nested concat expression in the select clause which "
        "uses a group variable gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_graph_patterns__having",
        'Test if "having" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_graph_patterns__join",
        'Test if "join" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_graph_patterns__left_join",
        'Test if "left join" gets properly translated into "OPTIONAL {...}" in the query text.',
    ),
    AlgebraTest(
        "test_graph_patterns__minus",
        'Test if "minus" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_graph_patterns__union",
        'Test if "union" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_integration__complex_query1",
        "Test a query with multiple graph patterns and solution modifiers"
        "gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_operators__arithmetics",
        "Test if arithmetics are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_operators__conditional_and",
        'Test if "conditional ands (&&)" are properly translated into the query text.',
    ),
    AlgebraTest(
        "test_operators__conditional_or",
        'Test if "conditional ors (||)" are properly translated into the query text.',
    ),
    AlgebraTest(
        "test_operators__relational",
        "Test if relational expressions are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_operators__unary",
        "Test if unary expressions are properly translated into the query text.",
    ),
    AlgebraTest(
        "test_other__service2",
        'Test if "service" along with its service string is properly translated'
        "into the query text.",
    ),
    AlgebraTest(
        "test_other__values",
        'Test if "values" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_property_path__alternative_path",
        "Test if an alternative path gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_property_path__inverse_path",
        "Test if an inverse path gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_property_path__negated_property_set",
        "Test if a negated property set gets properly translated into the query text.",
        pytest.mark.xfail(
            raises=TypeError, reason="fails with TypeError in translateAlgebra"
        ),
    ),
    AlgebraTest(
        "test_property_path__one_or_more_path",
        "Test if a oneOrMore path gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_property_path__sequence_path",
        "Test if a sequence path gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_property_path__zero_or_more_path",
        "Test if a zeroOrMore path gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_property_path__zero_or_one_path",
        "Test if a zeroOrOne path gets properly translated into the query text.",
    ),
    AlgebraTest(
        "test_solution_modifiers__distinct",
        'Test if "distinct" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_solution_modifiers__order_by",
        'Test if "order by" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_solution_modifiers__reduced",
        'Test if "reduced" gets properly translated into the query text.',
    ),
    AlgebraTest(
        "test_solution_modifiers__slice",
        "Test if slice get properly translated into the limit and offset.",
    ),
    AlgebraTest(
        "test_solution_modifiers__to_multiset",
        "Test if subqueries get properly translated into the query text.",
    ),
]

unused_files = {
    "test_property_path__predicate_path.txt",
    "test_solution_modifiers__project.txt",
}


if os.name != "nt":
    # On Windows this test causes a stackoverflow or memory error, so don't run
    # it on Windows. The test fails anyway on linux and should be fixed. Once
    # the cause of this failure is fixed the condition to not run it on windows
    # should be removed and it should be added to the normal tests.
    algebra_tests.append(
        AlgebraTest(
            "test_other__service1",
            "Test if a nested service pattern is properly translated"
            "into the query text.",
            pytest.mark.xfail(
                raises=RecursionError,
                reason="Fails with RecursionError inside parser.parseQuery",
            ),
        )
    )
else:
    unused_files.add("test_other__service1.txt")


def test_all_files_used(data_path: Path) -> None:
    all_files_names = {path.name for path in data_path.glob("*")}
    expected_files = {test.filename for test in algebra_tests}

    # These files are not being used, and they are empty, they were likely
    # added as placeholders.

    all_files_names.difference_update(unused_files)

    assert expected_files == all_files_names


@pytest.mark.parametrize("test_spec", [test.pytest_param() for test in algebra_tests])
def test_roundtrip(test_spec: AlgebraTest, data_path: Path) -> None:
    """
    Query remains the same over two successive parse and translate cycles.
    """
    query_text = (data_path / test_spec.filename).read_text()

    logging.info("checking expectation: %s", test_spec.description)
    query_tree = parser.parseQuery(query_text)
    query_algebra = algebra.translateQuery(query_tree)
    query_from_algebra = translateAlgebra(query_algebra)
    logging.debug(
        "query_from_query_from_algebra = \n%s",
        query_from_algebra,
    )

    query_tree_2 = parser.parseQuery(query_from_algebra)
    query_algebra_2 = algebra.translateQuery(query_tree_2)
    query_from_query_from_algebra = translateAlgebra(query_algebra_2)
    logging.debug(
        "query_from_query_from_algebra = \n%s",
        query_from_query_from_algebra,
    )
    assert (
        query_from_algebra == query_from_query_from_algebra
    ), f"failed expectation: {test_spec.description}"

    # TODO: Execute the raw query (query_text) and the reconstituted query
    # (query_from_query_from_algebra) against a well defined graph and ensure
    # they yield the same result.


def test_sparql_group_concat():
    """Tests if GROUP_CONCAT correctly uses the separator keyword"""
    query = """
    PREFIX : <http://example.org/>

    SELECT ?subject (GROUP_CONCAT(?object; separator="")
                        AS ?concatenatedObjects)
    WHERE {
      VALUES (?subject ?object) {
        (:pred "a")
        (:pred "b")
        (:pred "c")
      }
    }
    GROUP BY ?subject
    """

    g = Graph()
    q = dict(g.query(query))
    assert q[URIRef("http://example.org/pred")] == Literal("abc")


def test_sparql_blank_node_comma():
    """Tests if blank nodes separated by commas are correctly parsed"""

    query = """
    PREFIX : <http://example.org/>

    SELECT ?s WHERE {
    ?s :hasIngredient [:name "chicken"], [:name "butter"] .
    } LIMIT 10
    """

    parse_results = parser.parseQuery(query)
    triples = parse_results[1]["where"].part[0].triples[0]
    s_count = sum(1 for i in range(0, len(triples), 3) if triples[i] == Variable("s"))
    assert s_count == 2, f"Found ?s as subject {s_count} times, expected 2"



"""
Large parts of this module are taken from the ``isodate`` package.
https://pypi.org/project/isodate/
Modifications are made to isodate features to allow compatibility with
XSD dates and durations that are not necessarily valid ISO8601 strings.

Copyright (c) 2024, Ashley Sommer, and RDFLib contributors
Copyright (c) 2021, Hugo van Kemenade and contributors
Copyright (c) 2009-2018, Gerhard Weis and contributors
Copyright (c) 2009, Gerhard Weis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
- Neither the name of the <organization> nor the
names of its contributors may be used to endorse or promote products
derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

import re
import sys
from datetime import date, datetime, time, timedelta
from decimal import ROUND_FLOOR, Decimal
from typing import List, Tuple, Union, cast

if sys.version_info[:3] < (3, 11, 0):
    from isodate import parse_date, parse_datetime, parse_time
else:
    # On python 3.11, use the built-in parsers
    parse_date = date.fromisoformat
    parse_datetime = datetime.fromisoformat
    parse_time = time.fromisoformat


def fquotmod(
    val: Decimal, low: Union[Decimal, int], high: Union[Decimal, int]
) -> Tuple[int, Decimal]:
    """
    A divmod function with boundaries.

    """
    # assumes that all the maths is done with Decimals.
    # divmod for Decimal uses truncate instead of floor as builtin
    # divmod, so we have to do it manually here.
    a: Decimal = val - low
    b: Union[Decimal, int] = high - low
    div: Decimal = (a / b).to_integral(ROUND_FLOOR)
    mod: Decimal = a - div * b
    # if we were not using Decimal, it would look like this.
    # div, mod = divmod(val - low, high - low)
    mod += low
    return int(div), mod


def max_days_in_month(year: int, month: int) -> int:
    """
    Determines the number of days of a specific month in a specific year.
    """
    if month in (1, 3, 5, 7, 8, 10, 12):
        return 31
    if month in (4, 6, 9, 11):
        return 30
    if month < 1 or month > 12:
        raise ValueError("Month must be in 1..12")
    # Month is February
    if ((year % 400) == 0) or ((year % 100) != 0) and ((year % 4) == 0):
        return 29
    return 28


class Duration:
    """
    A class which represents a duration.

    The difference to datetime.timedelta is, that this class handles also
    differences given in years and months.
    A Duration treats differences given in year, months separately from all
    other components.

    A Duration can be used almost like any timedelta object, however there
    are some restrictions:
    - It is not really possible to compare Durations, because it is unclear,
    whether a duration of 1 year is bigger than 365 days or not.
    - Equality is only tested between the two (year, month vs. timedelta)
    basic components.

    A Duration can also be converted into a datetime object, but this requires
    a start date or an end date.

    The algorithm to add a duration to a date is defined at
    http://www.w3.org/TR/xmlschema-2/#adding-durations-to-dateTimes
    """

    def __init__(
        self,
        days: float = 0,
        seconds: float = 0,
        microseconds: float = 0,
        milliseconds: float = 0,
        minutes: float = 0,
        hours: float = 0,
        weeks: float = 0,
        months: Union[Decimal, float, int, str] = 0,
        years: Union[Decimal, float, int, str] = 0,
    ):
        """
        Initialise this Duration instance with the given parameters.
        """
        if not isinstance(months, Decimal):
            months = Decimal(str(months))
        if not isinstance(years, Decimal):
            years = Decimal(str(years))
        new_years, months = fquotmod(months, 0, 12)
        self.months = months
        self.years = Decimal(years + new_years)
        self.tdelta = timedelta(
            days, seconds, microseconds, milliseconds, minutes, hours, weeks
        )
        if self.years < 0 and self.tdelta.days < 0:
            raise ValueError("Duration cannot have negative years and negative days")

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, name):
        """
        Provide direct access to attributes of included timedelta instance.
        """
        return getattr(self.tdelta, name)

    def __str__(self):
        """
        Return a string representation of this duration similar to timedelta.
        """
        params = []
        if self.years:
            params.append("%d years" % self.years)
        if self.months:
            fmt = "%d months"
            if self.months <= 1:
                fmt = "%d month"
            params.append(fmt % self.months)
        params.append(str(self.tdelta))
        return ", ".join(params)

    def __repr__(self):
        """
        Return a string suitable for repr(x) calls.
        """
        return "%s.%s(%d, %d, %d, years=%s, months=%s)" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.tdelta.days,
            self.tdelta.seconds,
            self.tdelta.microseconds,
            str(self.years),
            str(self.months),
        )

    def __hash__(self):
        """
        Return a hash of this instance so that it can be used in, for
        example, dicts and sets.
        """
        return hash((self.tdelta, self.months, self.years))

    def __neg__(self):
        """
        A simple unary minus.

        Returns a new Duration instance with all it's negated.
        """
        negduration = Duration(years=-self.years, months=-self.months)
        negduration.tdelta = -self.tdelta
        return negduration

    def __add__(self, other: Union[Duration, timedelta, date, datetime]):
        """
        Durations can be added with Duration, timedelta, date and datetime
        objects.
        """
        if isinstance(other, Duration):
            newduration = Duration(
                years=self.years + other.years, months=self.months + other.months
            )
            newduration.tdelta = self.tdelta + other.tdelta
            return newduration
        elif isinstance(other, timedelta):
            newduration = Duration(years=self.years, months=self.months)
            newduration.tdelta = self.tdelta + other
            return newduration
        try:
            # try anything that looks like a date or datetime
            # 'other' has attributes year, month, day
            # and relies on 'timedelta + other' being implemented
            if not (float(self.years).is_integer() and float(self.months).is_integer()):
                raise ValueError(
                    "fractional years or months not supported for date calculations"
                )
            newmonth: Decimal = Decimal(other.month) + self.months
            carry, newmonth = fquotmod(newmonth, 1, 13)
            newyear: int = other.year + int(self.years) + carry
            maxdays: int = max_days_in_month(newyear, int(newmonth))
            newday: Union[int, float]
            if other.day > maxdays:
                newday = maxdays
            else:
                newday = other.day
            newdt = other.replace(year=newyear, month=int(newmonth), day=newday)
            # does a timedelta + date/datetime
            return self.tdelta + newdt
        except AttributeError:
            # other probably was not a date/datetime compatible object
            pass
        # we have tried everything .... return a NotImplemented
        return NotImplemented

    __radd__ = __add__

    def __mul__(self, other):
        if isinstance(other, int):
            newduration = Duration(years=self.years * other, months=self.months * other)
            newduration.tdelta = self.tdelta * other
            return newduration
        return NotImplemented

    __rmul__ = __mul__

    def __sub__(self, other: Union[Duration, timedelta]):
        """
        It is possible to subtract Duration and timedelta objects from Duration
        objects.
        """
        if isinstance(other, Duration):
            newduration = Duration(
                years=self.years - other.years, months=self.months - other.months
            )
            newduration.tdelta = self.tdelta - other.tdelta
            return newduration
        try:
            # do maths with our timedelta object ....
            newduration = Duration(years=self.years, months=self.months)
            newduration.tdelta = self.tdelta - other
            return newduration
        except TypeError:
            # looks like timedelta - other is not implemented
            pass
        return NotImplemented

    def __rsub__(self, other: Union[timedelta, date, datetime]):
        """
        It is possible to subtract Duration objects from date, datetime and
        timedelta objects.
        """
        # TODO: there is some weird behaviour in date - timedelta ...
        #       if timedelta has seconds or microseconds set, then
        #       date - timedelta != date + (-timedelta)
        #       for now we follow this behaviour to avoid surprises when mixing
        #       timedeltas with Durations, but in case this ever changes in
        #       the stdlib we can just do:
        #       return -self + other
        #       instead of all the current code

        if isinstance(other, timedelta):
            tmpdur = Duration()
            tmpdur.tdelta = other
            return tmpdur - self
        try:
            # check if other behaves like a date/datetime object
            # does it have year, month, day and replace?
            if not (float(self.years).is_integer() and float(self.months).is_integer()):
                raise ValueError(
                    "fractional years or months not supported for date calculations"
                )
            newmonth: Decimal = Decimal(other.month) - self.months
            carry, newmonth = fquotmod(newmonth, 1, 13)
            newyear: int = other.year - int(self.years) + carry
            maxdays: int = max_days_in_month(newyear, int(newmonth))
            newday: Union[int, float]
            if other.day > maxdays:
                newday = maxdays
            else:
                newday = other.day
            newdt = other.replace(year=newyear, month=int(newmonth), day=newday)
            return newdt - self.tdelta
        except AttributeError:
            # other probably was not compatible with data/datetime
            pass
        return NotImplemented

    def __eq__(self, other):
        """
        If the years, month part and the timedelta part are both equal, then
        the two Durations are considered equal.
        """
        if isinstance(other, Duration):
            if (self.years * 12 + self.months) == (
                other.years * 12 + other.months
            ) and self.tdelta == other.tdelta:
                return True
            return False
        # check if other con be compared against timedelta object
        # will raise an AssertionError when optimisation is off
        if self.years == 0 and self.months == 0:
            return self.tdelta == other
        return False

    def __ne__(self, other):
        """
        If the years, month part or the timedelta part is not equal, then
        the two Durations are considered not equal.
        """
        if isinstance(other, Duration):
            if (self.years * 12 + self.months) != (
                other.years * 12 + other.months
            ) or self.tdelta != other.tdelta:
                return True
            return False
        # check if other can be compared against timedelta object
        # will raise an AssertionError when optimisation is off
        if self.years == 0 and self.months == 0:
            return self.tdelta != other
        return True

    def totimedelta(self, start=None, end=None):
        """
        Convert this duration into a timedelta object.

        This method requires a start datetime or end datetime, but raises
        an exception if both are given.
        """
        if start is None and end is None:
            raise ValueError("start or end required")
        if start is not None and end is not None:
            raise ValueError("only start or end allowed")
        if start is not None:
            return (start + self) - start
        return end - (end - self)


ISO8601_PERIOD_REGEX = re.compile(
    r"^(?P<sign>[+-])?"
    r"P(?!\b)"
    r"(?P<years>[0-9]+([,.][0-9]+)?Y)?"
    r"(?P<months>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<weeks>[0-9]+([,.][0-9]+)?W)?"
    r"(?P<days>[0-9]+([,.][0-9]+)?D)?"
    r"((?P<separator>T)(?P<hours>[0-9]+([,.][0-9]+)?H)?"
    r"(?P<minutes>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<seconds>[0-9]+([,.][0-9]+)?S)?)?$"
)
# regular expression to parse ISO duration strings.


def parse_xsd_duration(
    dur_string: str, as_timedelta_if_possible: bool = True
) -> Union[Duration, timedelta]:
    """
    Parses an ISO 8601 durations into datetime.timedelta or Duration objects.

    If the ISO date string does not contain years or months, a timedelta
    instance is returned, else a Duration instance is returned.

    The following duration formats are supported:
      -``PnnW``                  duration in weeks
      -``PnnYnnMnnDTnnHnnMnnS``  complete duration specification
      -``PYYYYMMDDThhmmss``      basic alternative complete date format
      -``PYYYY-MM-DDThh:mm:ss``  extended alternative complete date format
      -``PYYYYDDDThhmmss``       basic alternative ordinal date format
      -``PYYYY-DDDThh:mm:ss``    extended alternative ordinal date format

    The '-' is optional.

    Limitations:  ISO standard defines some restrictions about where to use
    fractional numbers and which component and format combinations are
    allowed. This parser implementation ignores all those restrictions and
    returns something when it is able to find all necessary components.
    In detail:
    - it does not check, whether only the last component has fractions.
    - it allows weeks specified with all other combinations
    The alternative format does not support durations with years, months or
    days set to 0.
    """
    if not isinstance(dur_string, str):
        raise TypeError(f"Expecting a string: {dur_string!r}")
    match = ISO8601_PERIOD_REGEX.match(dur_string)
    if not match:
        # try alternative format:
        if dur_string.startswith("P"):
            durdt = parse_datetime(dur_string[1:])
            if as_timedelta_if_possible and durdt.year == 0 and durdt.month == 0:
                # FIXME: currently not possible in alternative format
                # create timedelta
                return timedelta(
                    days=durdt.day,
                    seconds=durdt.second,
                    microseconds=durdt.microsecond,
                    minutes=durdt.minute,
                    hours=durdt.hour,
                )
            else:
                # create Duration
                return Duration(
                    days=durdt.day,
                    seconds=durdt.second,
                    microseconds=durdt.microsecond,
                    minutes=durdt.minute,
                    hours=durdt.hour,
                    months=durdt.month,
                    years=durdt.year,
                )
        raise ValueError("Unable to parse duration string " + dur_string)
    groups = match.groupdict()
    for key, val in groups.items():
        if key not in ("separator", "sign"):
            if val is None:
                groups[key] = "0n"
            # print groups[key]
            if key in ("years", "months"):
                groups[key] = Decimal(groups[key][:-1].replace(",", "."))
            else:
                # these values are passed into a timedelta object,
                # which works with floats.
                groups[key] = float(groups[key][:-1].replace(",", "."))
    ret: Union[Duration, timedelta]
    if as_timedelta_if_possible and groups["years"] == 0 and groups["months"] == 0:
        ret = timedelta(
            days=groups["days"],  # type: ignore[arg-type]
            hours=groups["hours"],  # type: ignore[arg-type]
            minutes=groups["minutes"],  # type: ignore[arg-type]
            seconds=groups["seconds"],  # type: ignore[arg-type]
            weeks=groups["weeks"],  # type: ignore[arg-type]
        )
        if groups["sign"] == "-":
            ret = timedelta(0) - ret
    else:
        ret = Duration(
            years=cast(Decimal, groups["years"]),
            months=cast(Decimal, groups["months"]),
            days=groups["days"],  # type: ignore[arg-type]
            hours=groups["hours"],  # type: ignore[arg-type]
            minutes=groups["minutes"],  # type: ignore[arg-type]
            seconds=groups["seconds"],  # type: ignore[arg-type]
            weeks=groups["weeks"],  # type: ignore[arg-type]
        )
        if groups["sign"] == "-":
            ret = Duration(0) - ret

    return ret


def duration_isoformat(tdt: Union[Duration, timedelta], in_weeks: bool = False) -> str:
    if not in_weeks:
        ret: List[str] = []
        minus = False
        has_year_or_month = False
        if isinstance(tdt, Duration):
            if tdt.years == 0 and tdt.months == 0:
                pass  # don't do anything, we have no year or month
            else:
                has_year_or_month = True
                months = tdt.years * 12 + tdt.months
                if months < 0:
                    minus = True
                    months = abs(months)
                # We can use divmod instead of fquotmod here because its month_count
                # not month_index, and we don't have any negative months at this point.
                new_years, new_months = divmod(months, 12)
                if new_years:
                    ret.append(str(new_years) + "Y")
                if tdt.months:
                    ret.append(str(new_months) + "M")
            tdt = tdt.tdelta
        usecs: int = ((tdt.days * 86400) + tdt.seconds) * 1000000 + tdt.microseconds
        if usecs < 0:
            if minus:
                raise ValueError(
                    "Duration cannot have negative years and negative days"
                )
            elif has_year_or_month:
                raise ValueError(
                    "Duration cannot have positive years and months but negative days"
                )
            minus = True
            usecs = abs(usecs)
        if usecs == 0:
            # No delta parts other than years and months
            pass
        else:
            seconds, usecs = divmod(usecs, 1000000)
            minutes, seconds = divmod(seconds, 60)
            hours, minutes = divmod(minutes, 60)
            days, hours = divmod(hours, 24)
            if days:
                ret.append(str(days) + "D")
            if hours or minutes or seconds or usecs:
                ret.append("T")
                if hours:
                    ret.append(str(hours) + "H")
                if minutes:
                    ret.append(str(minutes) + "M")
                if seconds or usecs:
                    if usecs:
                        ret.append(("%d.%06d" % (seconds, usecs)).rstrip("0"))
                    else:
                        ret.append("%d" % seconds)
                    ret.append("S")
        if ret:
            return ("-P" if minus else "P") + "".join(ret)
        else:
            # at least one component has to be there.
            return "-P0D" if minus else "P0D"
    else:
        if tdt.days < 0:
            return f"-P{abs(tdt.days // 7)}W"
        return f"P{tdt.days // 7}W"


def xsd_datetime_isoformat(dt: datetime):
    if dt.microsecond == 0:
        no_tz_str = dt.strftime("%Y-%m-%dT%H:%M:%S")
    else:
        no_tz_str = dt.strftime("%Y-%m-%dT%H:%M:%S.%f")
    if dt.tzinfo is None:
        return no_tz_str
    else:
        offset_string = dt.strftime("%z")
        if offset_string == "+0000":
            return no_tz_str + "Z"
        first_char = offset_string[0]
        if first_char == "+" or first_char == "-":
            offset_string = offset_string[1:]
            sign = first_char
        else:
            sign = "+"
        tz_part = sign + offset_string[:2] + ":" + offset_string[2:]
        return no_tz_str + tz_part


def parse_xsd_date(date_string: str):
    """
    XSD Dates have more features than ISO8601 dates, specifically
    XSD allows timezones on dates, that must be stripped off.
    Also, XSD requires dashed separators, while ISO8601 is optional.
    RDFLib test suite has some date strings with times, the times are expected
    to be dropped during parsing.
    """
    if date_string.endswith("Z") or date_string.endswith("z"):
        date_string = date_string[:-1]
    if date_string.startswith("-"):
        date_string = date_string[1:]
        minus = True
    else:
        minus = False
    if "T" in date_string:
        # RDFLib test suite has some strange date strings, with times.
        # this has the side effect of also dropping the
        # TZ part, that is not wanted anyway for a date.
        date_string = date_string.split("T")[0]
    else:
        has_plus = date_string.rfind("+")
        if has_plus > 0:
            # Drop the +07:00 timezone part
            date_string = date_string[:has_plus]
        else:
            split_parts = date_string.rsplit("-", 1)
            if len(split_parts) > 1 and ":" in split_parts[-1]:
                # Drop the -09:00 timezone part
                date_string = split_parts[0]
    if "-" not in date_string:
        raise ValueError("XSD Date string must contain at least two dashes")
    return parse_date(date_string if not minus else ("-" + date_string))


def parse_xsd_gyear(gyear_string: str):
    """
    XSD gYear has more features than ISO8601 dates, specifically
    XSD allows timezones on a gYear, that must be stripped off.
    """
    if gyear_string.endswith("Z") or gyear_string.endswith("z"):
        gyear_string = gyear_string[:-1]
    if gyear_string.startswith("-"):
        gyear_string = gyear_string[1:]
        minus = True
    else:
        minus = False
    has_plus = gyear_string.rfind("+")
    if has_plus > 0:
        # Drop the +07:00 timezone part
        gyear_string = gyear_string[:has_plus]
    else:
        split_parts = gyear_string.rsplit("-", 1)
        if len(split_parts) > 1 and ":" in split_parts[-1]:
            # Drop the -09:00 timezone part
            gyear_string = split_parts[0]
    if len(gyear_string) < 4:
        raise ValueError("gYear string must be at least 4 numerals in length")
    gyear_string = gyear_string.lstrip("0")  # strip all leading zeros
    try:
        y = int(gyear_string if not minus else ("-" + gyear_string))
    except ValueError:
        raise ValueError("gYear string must be a valid integer")
    return date(y, 1, 1)


def parse_xsd_gyearmonth(gym_string: str):
    """
    XSD gYearMonth has more features than ISO8601 dates, specifically
    XSD allows timezones on a gYearMonth, that must be stripped off.
    """
    if gym_string.endswith("Z") or gym_string.endswith("z"):
        gym_string = gym_string[:-1]
    if gym_string.startswith("-"):
        gym_string = gym_string[1:]
        minus = True
    else:
        minus = False
    has_plus = gym_string.rfind("+")
    if has_plus > 0:
        # Drop the +07:00 timezone part
        gym_string = gym_string[:has_plus]
    else:
        split_parts = gym_string.rsplit("-", 1)
        if len(split_parts) > 1 and ":" in split_parts[-1]:
            # Drop the -09:00 timezone part
            gym_string = split_parts[0]
    year_month_parts = gym_string.split("-", 1)
    if len(year_month_parts) < 2:
        raise ValueError("XSD gYearMonth string must contain one dash")

    if len(year_month_parts[0]) < 4:
        raise ValueError("gYearMonth Year part must be at least 4 numerals in length")
    elif len(year_month_parts[1]) < 2:
        raise ValueError("gYearMonth Month part must be exactly 2 numerals in length")
    year_string = year_month_parts[0].lstrip("0")  # strip all leading zeros
    month_string = year_month_parts[1].lstrip("0")  # strip all leading zeros
    try:
        y = int(year_string if not minus else ("-" + year_string))
    except ValueError:
        raise ValueError("gYearMonth Year part must be a valid integer")
    try:
        m = int(month_string)
    except ValueError:
        raise ValueError("gYearMonth Month part must be a valid integer")
    return date(y, m, 1)


# Parse XSD Datetime is the same as ISO8601 Datetime
# It uses datetime.fromisoformat for python 3.11 and above
# or isodate.parse_datetime for older versions
# parse_xsd_datetime = parse_datetime

# Parse XSD Time is the same as ISO8601 Time
# It uses time.fromisoformat for python 3.11 and above
# or isodate.parse_time for older versions
# parse_xsd_time = parse_time



import pytest

from rdflib import OWL, RDF, RDFS, Graph, Namespace
from rdflib.extras.infixowl import ClassNamespaceFactory, Individual, Property, some

EXNS = Namespace("http://example.org/vocab/")

EXCL = ClassNamespaceFactory(EXNS)


@pytest.fixture(scope="function")
def graph():
    g = Graph(identifier=EXNS.context0)
    g.bind("ex", EXNS)
    Individual.factoryGraph = g

    yield g

    del g


def test_logic_structuring(graph):
    isPartOf = Property(EXNS.isPartOf)  # noqa: N806
    graph.add((isPartOf.identifier, RDF.type, OWL.TransitiveProperty))

    hasLocation = Property(EXNS.hasLocation, subPropertyOf=[isPartOf])  # noqa: N806
    graph.add((hasLocation.identifier, RDFS.subPropertyOf, isPartOf.identifier))

    leg = EXCL.Leg
    knee = EXCL.Knee
    joint = EXCL.Joint
    kneeJoint = EXCL.KneeJoint  # noqa: N806
    legStructure = EXCL.LegStructure  # noqa: N806

    kneeJoint.equivalentClass = [joint & (isPartOf @ some @ knee)]

    structure = EXCL.Structure

    legStructure.equivalentClass = [structure & (isPartOf @ some @ leg)]

    structure += leg
    structure += joint

    locatedInLeg = hasLocation @ some @ leg  # noqa: N806
    locatedInLeg += knee  # noqa: N806

    assert graph.serialize(format="ttl") == (
        "@prefix ex: <http://example.org/vocab/> .\n"
        "@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
        "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
        "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n"
        "\n"
        "ex:KneeJoint a owl:Class ;\n"
        "    owl:equivalentClass [ a owl:Class ;\n"
        "            owl:intersectionOf ( ex:Joint [ a owl:Restriction ;\n"
        "                        owl:onProperty ex:isPartOf ;\n"
        "                        owl:someValuesFrom ex:Knee ] ) ] .\n"
        "\n"
        "ex:LegStructure a owl:Class ;\n"
        "    owl:equivalentClass [ a owl:Class ;\n"
        "            owl:intersectionOf ( ex:Structure [ a owl:Restriction ;\n"
        "                        owl:onProperty ex:isPartOf ;\n"
        "                        owl:someValuesFrom ex:Leg ] ) ] .\n"
        "\n"
        "ex:Joint a owl:Class ;\n"
        "    rdfs:subClassOf ex:Structure .\n"
        "\n"
        "ex:Knee a owl:Class ;\n"
        "    rdfs:subClassOf [ a owl:Restriction ;\n"
        "            owl:onProperty ex:hasLocation ;\n"
        "            owl:someValuesFrom ex:Leg ] .\n"
        "\n"
        "ex:hasLocation a owl:ObjectProperty ;\n"
        "    rdfs:subPropertyOf ex:isPartOf .\n"
        "\n"
        "ex:Leg a owl:Class ;\n"
        "    rdfs:subClassOf ex:Structure .\n"
        "\n"
        "ex:Structure a owl:Class .\n"
        "\n"
        "ex:isPartOf a owl:ObjectProperty,\n"
        "        owl:TransitiveProperty .\n"
        "\n"
    )



from __future__ import annotations

import enum
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path, PurePath
from typing import IO, Callable, Generator, Optional, TextIO, Union

DestParmType = Union[Path, PurePath, str, IO[bytes], TextIO]


@dataclass(frozen=True)
class DestRef:
    param: DestParmType
    path: Path


class DestinationType(str, enum.Enum):
    RETURN = enum.auto()
    PATH = enum.auto()
    PURE_PATH = enum.auto()
    STR_PATH = enum.auto()
    FILE_URI = enum.auto()
    BINARY_IO = enum.auto()
    TEXT_IO = enum.auto()

    @contextmanager
    def make_ref(
        self,
        tmp_path: Path,
        encoding: Optional[str] = None,
        path_factory: Callable[[Path, DestinationType, Optional[str]], Path] = (
            lambda tmp_path, type, encoding: tmp_path / f"file-{type.name}-{encoding}"
        ),
    ) -> Generator[Optional[DestRef], None, None]:
        path = path_factory(tmp_path, self, encoding)
        # path = tmp_path / f"file-{self.name}"
        if self is DestinationType.RETURN:
            yield None
        elif self is DestinationType.PATH:
            yield DestRef(path, path)
        elif self is DestinationType.PURE_PATH:
            yield DestRef(PurePath(path), path)
        elif self is DestinationType.STR_PATH:
            yield DestRef(f"{path}", path)
        elif self is DestinationType.FILE_URI:
            yield DestRef(path.as_uri(), path)
        elif self is DestinationType.BINARY_IO:
            with path.open("wb") as bfh:
                yield DestRef(bfh, path)
        elif self is DestinationType.TEXT_IO:
            assert encoding is not None
            with path.open("w", encoding=encoding) as fh:
                yield DestRef(fh, path)
        else:
            raise ValueError(f"unsupported type {type!r}")



"""

NOTE: PyParsing setResultName/__call__ provides a very similar solution to this
I didn't realise at the time of writing and I will remove a
lot of this code at some point

Utility classes for creating an abstract-syntax tree out with pyparsing actions

Lets you label and group parts of parser production rules

For example:

# [5] BaseDecl ::= 'BASE' IRIREF
BaseDecl = Comp('Base', Keyword('BASE') + Param('iri',IRIREF))

After parsing, this gives you back an CompValue object,
which is a dict/object with the parameters specified.
So you can access the parameters are attributes or as keys:

baseDecl.iri

Comp lets you set an evalFn that is bound to the eval method of
the resulting CompValue


"""

from __future__ import annotations

from collections import OrderedDict
from types import MethodType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from pyparsing import ParserElement, ParseResults, TokenConverter, originalTextFor

from rdflib.term import BNode, Identifier, Variable

if TYPE_CHECKING:
    from rdflib.plugins.sparql.sparql import FrozenBindings


# This is an alternative

# Comp('Sum')( Param('x')(Number) + '+' + Param('y')(Number) )


def value(
    ctx: FrozenBindings,
    val: Any,
    variables: bool = False,
    errors: bool = False,
) -> Any:
    """
    utility function for evaluating something...

    Variables will be looked up in the context
    Normally, non-bound vars is an error,
    set variables=True to return unbound vars

    Normally, an error raises the error,
    set errors=True to return error

    """

    if isinstance(val, Expr):
        return val.eval(ctx)  # recurse?
    elif isinstance(val, CompValue):
        raise Exception("What do I do with this CompValue? %s" % val)

    elif isinstance(val, list):
        return [value(ctx, x, variables, errors) for x in val]

    elif isinstance(val, (BNode, Variable)):
        r = ctx.get(val)
        if isinstance(r, SPARQLError) and not errors:
            raise r
        if r is not None:
            return r

        # not bound
        if variables:
            return val
        else:
            raise NotBoundError

    elif isinstance(val, ParseResults) and len(val) == 1:
        return value(ctx, val[0], variables, errors)
    else:
        return val


class ParamValue:
    """
    The result of parsing a Param
    This just keeps the name/value
    All cleverness is in the CompValue
    """

    def __init__(
        self, name: str, tokenList: Union[List[Any], ParseResults], isList: bool
    ):
        self.isList = isList
        self.name = name
        if isinstance(tokenList, (list, ParseResults)) and len(tokenList) == 1:
            tokenList = tokenList[0]

        self.tokenList = tokenList

    def __str__(self) -> str:
        return "Param(%s, %s)" % (self.name, self.tokenList)


class Param(TokenConverter):
    """
    A pyparsing token for labelling a part of the parse-tree
    if isList is true repeat occurrences of ParamList have
    their values merged in a list
    """

    def __init__(self, name: str, expr, isList: bool = False):
        self.isList = isList
        TokenConverter.__init__(self, expr)
        self.setName(name)
        self.addParseAction(self.postParse2)

    def postParse2(self, tokenList: Union[List[Any], ParseResults]) -> ParamValue:
        return ParamValue(self.name, tokenList, self.isList)


class ParamList(Param):
    """
    A shortcut for a Param with isList=True
    """

    def __init__(self, name: str, expr):
        Param.__init__(self, name, expr, True)


_ValT = TypeVar("_ValT")


class CompValue(OrderedDict):
    """
    The result of parsing a Comp
    Any included Params are available as Dict keys
    or as attributes

    """

    def __init__(self, name: str, **values):
        OrderedDict.__init__(self)
        self.name = name
        self.update(values)

    def clone(self) -> CompValue:
        return CompValue(self.name, **self)

    def __str__(self) -> str:
        return self.name + "_" + OrderedDict.__str__(self)

    def __repr__(self) -> str:
        return self.name + "_" + dict.__repr__(self)

    def _value(
        self, val: _ValT, variables: bool = False, errors: bool = False
    ) -> Union[_ValT, Any]:
        if self.ctx is not None:
            return value(self.ctx, val, variables)
        else:
            return val

    def __getitem__(self, a):
        return self._value(OrderedDict.__getitem__(self, a))

    # type error: Signature of "get" incompatible with supertype "dict"
    # type error: Signature of "get" incompatible with supertype "Mapping"  [override]
    def get(self, a, variables: bool = False, errors: bool = False):  # type: ignore[override]
        return self._value(OrderedDict.get(self, a, a), variables, errors)

    def __getattr__(self, a: str) -> Any:
        # Hack hack: OrderedDict relies on this
        if a in ("_OrderedDict__root", "_OrderedDict__end"):
            raise AttributeError()
        try:
            return self[a]
        except KeyError:
            # raise AttributeError('no such attribute '+a)
            return None

    if TYPE_CHECKING:
        # this is here because properties are dynamically set on CompValue
        def __setattr__(self, __name: str, __value: Any) -> None: ...


class Expr(CompValue):
    """
    A CompValue that is evaluatable
    """

    def __init__(
        self,
        name: str,
        evalfn: Optional[Callable[[Any, Any], Any]] = None,
        **values,
    ):
        super(Expr, self).__init__(name, **values)

        self._evalfn = None
        if evalfn:
            self._evalfn = MethodType(evalfn, self)

    def eval(self, ctx: Any = {}) -> Union[SPARQLError, Any]:
        try:
            self.ctx: Optional[Union[Mapping, FrozenBindings]] = ctx
            # type error: "None" not callable
            return self._evalfn(ctx)  # type: ignore[misc]
        except SPARQLError as e:
            return e
        finally:
            self.ctx = None


class Comp(TokenConverter):
    """
    A pyparsing token for grouping together things with a label
    Any sub-tokens that are not Params will be ignored.

    Returns CompValue / Expr objects - depending on whether evalFn is set.
    """

    def __init__(self, name: str, expr: ParserElement):
        self.expr = expr
        TokenConverter.__init__(self, expr)
        self.setName(name)
        self.evalfn: Optional[Callable[[Any, Any], Any]] = None

    def postParse(
        self, instring: str, loc: int, tokenList: ParseResults
    ) -> Union[Expr, CompValue]:
        res: Union[Expr, CompValue]
        if self.evalfn:
            res = Expr(self.name)
            res._evalfn = MethodType(self.evalfn, res)
        else:
            res = CompValue(self.name)
            if self.name == "ServiceGraphPattern":
                # Then this must be a service graph pattern and have
                # already matched.
                # lets assume there is one, for now, then test for two later.
                sgp = originalTextFor(self.expr)
                service_string = sgp.searchString(instring)[0][0]
                res["service_string"] = service_string

        for t in tokenList:
            if isinstance(t, ParamValue):
                if t.isList:
                    if t.name not in res:
                        res[t.name] = []
                    res[t.name].append(t.tokenList)
                else:
                    res[t.name] = t.tokenList
                # res.append(t.tokenList)
            # if isinstance(t,CompValue):
            #    res.update(t)
        return res

    def setEvalFn(self, evalfn: Callable[[Any, Any], Any]) -> Comp:
        self.evalfn = evalfn
        return self


def prettify_parsetree(t: ParseResults, indent: str = "", depth: int = 0) -> str:
    out: List[str] = []
    for e in t.asList():
        out.append(_prettify_sub_parsetree(e, indent, depth + 1))
    for k, v in sorted(t.items()):
        out.append("%s%s- %s:\n" % (indent, "  " * depth, k))
        out.append(_prettify_sub_parsetree(v, indent, depth + 1))
    return "".join(out)


def _prettify_sub_parsetree(
    t: Union[Identifier, CompValue, set, list, dict, Tuple, bool, None],
    indent: str = "",
    depth: int = 0,
) -> str:
    out: List[str] = []
    if isinstance(t, CompValue):
        out.append("%s%s> %s:\n" % (indent, "  " * depth, t.name))
        for k, v in t.items():
            out.append("%s%s- %s:\n" % (indent, "  " * (depth + 1), k))
            out.append(_prettify_sub_parsetree(v, indent, depth + 2))
    elif isinstance(t, dict):
        for k, v in t.items():
            out.append("%s%s- %s:\n" % (indent, "  " * (depth + 1), k))
            out.append(_prettify_sub_parsetree(v, indent, depth + 2))
    elif isinstance(t, list):
        for e in t:
            out.append(_prettify_sub_parsetree(e, indent, depth + 1))
    else:
        out.append("%s%s- %r\n" % (indent, "  " * depth, t))
    return "".join(out)


# hurrah for circular imports
from rdflib.plugins.sparql.sparql import NotBoundError, SPARQLError  # noqa: E402



from rdflib import URIRef
from rdflib.paths import Path

uri_tplt = "http://example.org/%s"


def test_path_div_future():
    path = URIRef(uri_tplt % "one") / URIRef(uri_tplt % "other")
    assert isinstance(path, Path)



from functools import partial

import rdflib.plugins.sparql.parser as p
from rdflib import Literal, Variable
from rdflib.plugins.sparql.algebra import translatePName, traverse
from rdflib.plugins.sparql.operators import simplify
from rdflib.plugins.sparql.sparql import Prologue, QueryContext, SPARQLError
from test.utils import eq_ as eq


def _eval(e, ctx=None):
    if not ctx:
        ctx = QueryContext()
    try:
        r = e.eval(ctx)
        if isinstance(r, SPARQLError):
            print(r)
            return False
        return r
    except SPARQLError:
        return False


def _translate(e):
    return simplify(traverse(e, visitPost=partial(translatePName, prologue=Prologue())))


def test_regex():
    assert _eval(_translate(p.Expression.parseString('REGEX("zxcabczxc","abc")')[0]))

    eq(
        bool(_eval(_translate(p.Expression.parseString('REGEX("zxczxc","abc")')[0]))),
        False,
    )

    assert _eval(_translate(p.Expression.parseString('REGEX("bbbaaaaabbb","ba*b")')[0]))


def test_arithmetic():
    eq(_eval(_translate(p.Expression.parseString("2+3")[0])).value, 5)
    eq(_eval(_translate(p.Expression.parseString("3-2")[0])).value, 1)

    eq(_eval(_translate(p.Expression.parseString("2*3")[0])).value, 6)
    eq(_eval(_translate(p.Expression.parseString("4/2")[0])).value, 2)

    eq(_eval(_translate(p.Expression.parseString("2+2+2")[0])).value, 6)
    eq(_eval(_translate(p.Expression.parseString("2-2+2")[0])).value, 2)
    eq(_eval(_translate(p.Expression.parseString("(2-2)+2")[0])).value, 2)
    eq(_eval(_translate(p.Expression.parseString("2-(2+2)")[0])).value, -2)

    eq(_eval(_translate(p.Expression.parseString("2*2*2")[0])).value, 8)
    eq(_eval(_translate(p.Expression.parseString("4/2*2")[0])).value, 4)
    eq(_eval(_translate(p.Expression.parseString("8/4*2")[0])).value, 4)
    eq(_eval(_translate(p.Expression.parseString("8/(4*2)")[0])).value, 1)
    eq(_eval(_translate(p.Expression.parseString("(2/2)*2")[0])).value, 2)
    eq(_eval(_translate(p.Expression.parseString("4/(2*2)")[0])).value, 1)

    eq(_eval(_translate(p.Expression.parseString("2+3*2")[0])).value, 8)
    eq(_eval(_translate(p.Expression.parseString("(2+3)*2")[0])).value, 10)
    eq(_eval(_translate(p.Expression.parseString("2+4/2")[0])).value, 4)
    eq(_eval(_translate(p.Expression.parseString("(2+4)/2")[0])).value, 3)


def test_arithmetic_var():
    ctx = QueryContext()
    ctx[Variable("x")] = Literal(2)

    eq(_eval(_translate(p.Expression.parseString("2+?x")[0]), ctx).value, 4)

    eq(_eval(_translate(p.Expression.parseString("?x+3")[0]), ctx).value, 5)
    eq(_eval(_translate(p.Expression.parseString("3-?x")[0]), ctx).value, 1)

    eq(_eval(_translate(p.Expression.parseString("?x*3")[0]), ctx).value, 6)
    eq(_eval(_translate(p.Expression.parseString("4/?x")[0]), ctx).value, 2)

    eq(_eval(_translate(p.Expression.parseString("?x+?x+?x")[0]), ctx).value, 6)
    eq(_eval(_translate(p.Expression.parseString("?x-?x+?x")[0]), ctx).value, 2)
    eq(_eval(_translate(p.Expression.parseString("(?x-?x)+?x")[0]), ctx).value, 2)
    eq(_eval(_translate(p.Expression.parseString("?x-(?x+?x)")[0]), ctx).value, -2)

    eq(_eval(_translate(p.Expression.parseString("?x*?x*?x")[0]), ctx).value, 8)
    eq(_eval(_translate(p.Expression.parseString("4/?x*?x")[0]), ctx).value, 4)
    eq(_eval(_translate(p.Expression.parseString("8/4*?x")[0]), ctx).value, 4)
    eq(_eval(_translate(p.Expression.parseString("8/(4*?x)")[0]), ctx).value, 1)
    eq(_eval(_translate(p.Expression.parseString("(?x/?x)*?x")[0]), ctx).value, 2)
    eq(_eval(_translate(p.Expression.parseString("4/(?x*?x)")[0]), ctx).value, 1)


def test_comparisons():
    eq(bool(_eval(_translate(p.Expression.parseString("2<3")[0]))), True)
    eq(bool(_eval(_translate(p.Expression.parseString("2<3.0")[0]))), True)
    eq(bool(_eval(_translate(p.Expression.parseString("2<3e0")[0]))), True)

    eq(bool(_eval(_translate(p.Expression.parseString("4<3")[0]))), False)
    eq(bool(_eval(_translate(p.Expression.parseString("4<3.0")[0]))), False)
    eq(bool(_eval(_translate(p.Expression.parseString("4<3e0")[0]))), False)

    eq(bool(_eval(_translate(p.Expression.parseString("2<2.1")[0]))), True)
    eq(bool(_eval(_translate(p.Expression.parseString("2<21e-1")[0]))), True)

    eq(bool(_eval(_translate(p.Expression.parseString("2=2.0")[0]))), True)
    eq(bool(_eval(_translate(p.Expression.parseString("2=2e0")[0]))), True)

    eq(bool(_eval(_translate(p.Expression.parseString('2="cake"')[0]))), False)


def test_comparisons_var():
    ctx = QueryContext()
    ctx[Variable("x")] = Literal(2)

    eq(bool(_eval(_translate(p.Expression.parseString("?x<3")[0]), ctx)), True)
    eq(bool(_eval(_translate(p.Expression.parseString("?x<3.0")[0]), ctx)), True)
    eq(bool(_eval(_translate(p.Expression.parseString("?x<3e0")[0]), ctx)), True)

    eq(bool(_eval(_translate(p.Expression.parseString("?x<2.1")[0]), ctx)), True)
    eq(bool(_eval(_translate(p.Expression.parseString("?x<21e-1")[0]), ctx)), True)

    eq(bool(_eval(_translate(p.Expression.parseString("?x=2.0")[0]), ctx)), True)
    eq(bool(_eval(_translate(p.Expression.parseString("?x=2e0")[0]), ctx)), True)

    eq(bool(_eval(_translate(p.Expression.parseString('?x="cake"')[0]), ctx)), False)

    ctx = QueryContext()
    ctx[Variable("x")] = Literal(4)

    eq(bool(_eval(_translate(p.Expression.parseString("?x<3")[0]), ctx)), False)
    eq(bool(_eval(_translate(p.Expression.parseString("?x<3.0")[0]), ctx)), False)
    eq(bool(_eval(_translate(p.Expression.parseString("?x<3e0")[0]), ctx)), False)


def test_and_or():
    eq(bool(_eval(_translate(p.Expression.parseString("3>2 && 3>1")[0]))), True)
    eq(
        bool(_eval(_translate(p.Expression.parseString("3>2 && 3>4 || 2>1")[0]))),
        True,
    )
    eq(
        bool(_eval(_translate(p.Expression.parseString("2>1 || 3>2 && 3>4")[0]))),
        True,
    )
    eq(
        bool(_eval(_translate(p.Expression.parseString("(2>1 || 3>2) && 3>4")[0]))),
        False,
    )
